/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UNDOSYSTEM__
#define __UNDOSYSTEM__

#include "Uti3DCore.h"
#include "UtiType.h"

#include <map>
#include <set>


#define UNDO_USECHANGE
#define UNDO_FLAG true

class CUndoDocument;

class CUndoCommand : public ReferenceUtiOB
{
public:
   CUndoCommand();
   ~CUndoCommand();
   virtual void undo()=0;
   virtual void redo()=0;
   virtual void* getOwner() {return NULL;};
};

class CUndoSnapShot : public ReferenceUtiOB
{
protected:
   CReferenceList<CUndoCommand> m_commands;
   std::set<void*> m_commandset;
public:
   CUndoSnapShot();
   ~CUndoSnapShot();
   virtual void addCommand(CUndoCommand* command);
   virtual void undo();
   virtual void redo();
   virtual void rollback();
   virtual bool alreadyInThisSnapShot(void* data);
   virtual int commandCount();
};

class CUndoDocumentListener : public ReferenceUtiOB
{
public:
  CUndoDocumentListener();
  virtual void docChanged(CUndoDocument* doc)=0;
};

class CUndoDocument : public ReferenceUtiOB
{
protected:
   CReferenceList<CUndoSnapShot> m_undosnapshot;
   CReferenceList<CUndoSnapShot> m_redosnapshot;
   CUndoDocumentListener* m_listener;
   ReferenceUtiOB* m_document;
public:
   CUndoDocument();
   ~CUndoDocument();
   CUndoDocumentListener* getListener();
   void setListener(CUndoDocumentListener* doclistener);
   virtual void setDocument(ReferenceUtiOB* obj);
   virtual ReferenceUtiOB* getDocument();
   virtual void snapshot();
   virtual void rollback();
   virtual void undo();
   virtual void redo();
   virtual void addCommand(CUndoCommand* command);
   virtual bool alreadyInThisSnapShot(void* data);
   virtual void clearIfSnapShotIsEmpty();
};

class CUndoListener : public ReferenceUtiOB
{
  public:
    CUndoListener(); 
    virtual void undo_delete(void* watchedvalue, ReferenceUtiOB* changedata)=0;
};

class CUndoChangeListener : public CUndoListener
{
public:
   CUndoChangeListener();
   virtual void undo_prechange(void* watchedvalue, ReferenceUtiOB* changedata)=0;
   virtual void undo_postchange(void* watchedvalue, ReferenceUtiOB* changedata)=0;
  
};

struct CUndoPositionInfo
{
  int m_position;
  int m_position2;
};

class CUndoVectorListener : public CUndoListener
{
  public:
    CUndoVectorListener();
    virtual void undo_addvalue(void* watchedvalue, CUndoPositionInfo* position, ReferenceUtiOB* changedata)=0;
    virtual void undo_exchangevalue(void* watchedvalue, CUndoPositionInfo* position, ReferenceUtiOB* changedata)=0;
    virtual void undo_movevalue(void* watchedvalue, CUndoPositionInfo* position, ReferenceUtiOB* changedata)=0;
    virtual void undo_removevalue(void* watchedvalue, CUndoPositionInfo* position, ReferenceUtiOB* changedata)=0;
    virtual void undo_clearvalue(void* watchedvalue, ReferenceUtiOB* changedata)=0;
};

template<class T> class CWatchedUndo : public CUndoCommand
{
protected:
    T m_item;
    void* m_ptr;
public:
    CWatchedUndo(void* p, const T& value)
    {
       m_ptr = p;
       m_item = value;
    }
    ~CWatchedUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_ptr; 
    }
};

template <class T, bool t_doundo=UNDO_FLAG> class CWatchedValue {
protected:
friend class CWatchedUndo<T>;
      T m_data;
      bool m_initialised;
public:
    CWatchedValue()
    {
        m_data = T();
        m_initialised =false;
    };
    CWatchedValue(const T &a)
    {
         m_data = a;
         m_initialised = true;
    };
    ~CWatchedValue();
    inline T& getInternData() {
      return m_data;
    };
    void removeListeners();
    void addChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata);
    void removeChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata);
    inline void set(const T& a);
    inline T get() const
    {
       return m_data;
    }
    inline operator T() const
    {
        return m_data;
    }
    bool operator ==(const T& b) const
    {
        return m_data == b;
    }
    inline CWatchedValue& operator=(const T &a) throw()
    {
      #ifdef UNDO_USECHANGE
           set(a);
	   #else
	   if (t_doundo) {
	     set(a);
	   } else {
	     m_data = a;
	   };
	   #endif
        return *this;
    };
};

template<class T> void CWatchedUndo<T>::undo()
{
  CHECKEXISTANCE;
   CWatchedValue<T, true>* watched = (CWatchedValue<T, true>*)m_ptr;
   T temp = watched->m_data;
   watched->m_data = m_item;
   m_item = temp;
}

template<class T> void CWatchedUndo<T>::redo()
{CHECKEXISTANCE;
    CWatchedValue<T, true>* watched = (CWatchedValue<T, true>*)m_ptr;
   T temp = watched->m_data;
   watched->m_data = m_item;
   m_item = temp;
}


template<class T> class CWatchedReferenceUndo : public CUndoCommand
{
protected:
    uti_refptr<T> m_item;
    void* m_ptr;
public:
    CWatchedReferenceUndo(void* p, T* value)
    {
       m_ptr = p;
       m_item = value;
    }
    ~CWatchedReferenceUndo()
    {
    
    }
    virtual void undo();
    virtual void redo();
    void* getOwner()
    {
       return m_ptr; 
    }
};

template <class T, bool t_doundo=UNDO_FLAG> class CWatchedReference {
protected:
friend class CWatchedReferenceUndo<T>;
      uti_refptr<T> m_data;
public:
    CWatchedReference()
    {
    
    };
    CWatchedReference(const T &a)
    {
         m_data = a;     
    };
    ~CWatchedReference()
    {
    #ifdef UNDO_USECHANGE
      removeListeners();
       #endif
    }
    inline uti_refptr<T>& getInternData() {
      return m_data;
    };
    void removeListeners();
    void addChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata);
    void removeChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata);
    inline void set(T* a);
    inline T* get() const
    {
       return m_data.get();
    }
    inline operator T() const
    {
        return m_data.get();
    }
    bool operator ==(const T* b) const
    {
        return m_data.get() == b;
    }
    inline CWatchedReference& operator=(T *a) throw()
    {
      #ifdef UNDO_USECHANGE
           set(a);
	   #else
	   if (t_doundo) {
	     set(a);
	   } else {
	     m_data = a;
	   };
	   #endif
        return *this;
    };
    inline T* operator->() const throw() {return get();};
    inline T& operator*() const throw() {return *get();};
};

template<class T> void CWatchedReferenceUndo<T>::undo()
{
   CWatchedReference<T, true>* watched = (CWatchedReference<T, true>*)m_ptr;
   uti_refptr<T> temp = watched->m_data;
   watched->m_data = m_item;
   m_item = temp;
}

template<class T> void CWatchedReferenceUndo<T>::redo()
{
    CWatchedReference<T, true>* watched = (CWatchedReference<T, true>*)m_ptr;
   uti_refptr<T> temp = watched->m_data;
   watched->m_data = m_item;
   m_item = temp;
}


template<class T> class CWatchedReferenceAdapter : public CBaseReferenceAdapter
{
  protected:
    CWatchedReference<T>& m_ref;
  public:
    CWatchedReferenceAdapter(CWatchedReference<T>& ref) : m_ref(ref)
    {
      
    };
    
     virtual ReferenceUtiOB* get() const
     {
         return m_ref.get(); 
     }
    virtual void set(ReferenceUtiOB* ob)
    {
        m_ref = (T*)ob;
    };
    
};

struct CChangeListElement
{
  uti_refptr<ReferenceUtiOB> m_element;
  CChangeListElement() {
    
  };
  CChangeListElement(ReferenceUtiOB* obj) {
    m_element = obj;
  };
};

struct CUndoChangeList {
   std::map<CUndoListener*, CChangeListElement > m_updateobjects;
};

const int UNDO_HASMAPSIZE = 4481; //prim



class CUndoControl : public CUtiOB
{
protected:
    static THREADLOCAL CUndoControl* s_instance;
    std::map<void*, CUndoChangeList> m_changelist;
    Array<int> m_hashMap;
    void addToHashMap(void* th);
    void removeFromHashMap(void* th);
    inline bool isInHashMap(void* th)
    {
         unsigned int value = ((UTI_INT64)th) % UNDO_HASMAPSIZE;
         return m_hashMap[value] != 0;
    };
    uti_refptr<CUndoDocument> m_currentdoc;
    int m_doclockcount;
    int m_commandlockcount;
public:
    inline bool hasListener(void* th)
    {
       return isInHashMap(th);
    }
    void vec_clearvalue(void* th);
    void vec_addvalue(void* th, int position);
    void vec_removevalue(void* th, int position);
    void vec_movevalue(void* th, int m1, int m2);
    void vec_exchangevalue(void* th, int position);
    void preChange(void* th);
    void postChange(void* th);
    void deleteWatch(void* th);
    void addChangeListener(void* th, CUndoListener* listener, ReferenceUtiOB* changedata);
    void removeChangeListener(void* th, CUndoListener* listener, ReferenceUtiOB* changedata);
public:
   static CUndoControl* getInstance();
   CUndoControl();
   ~CUndoControl();
  public:
   void setDocument(CUndoDocument* document);
   
  public:
    CUndoDocument* getDocument();
   void lockDocument(CUndoDocument* doc);
   void unlockDocument(CUndoDocument*doc);
   void beginCommand();
   void endCommand();
   void snapshot();
   void rollback();
   void undo();
   void redo();
   void addUndoCommand(CUndoCommand* command);
   bool alreadyInThisSnapShot(void* data);
};

class CUndoDocumentLocker {
public:
  CUndoDocument* m_doc;
  explicit CUndoDocumentLocker(CUndoDocument* doc);
 ~CUndoDocumentLocker();
};

class CUndoCommandLocker {
 //  CUndoDocument* m_doc;
public:
  explicit CUndoCommandLocker();
 ~CUndoCommandLocker();
};

class CUndoCommandValueLocker {
public:
   bool m_locked;
   explicit CUndoCommandValueLocker();
 ~CUndoCommandValueLocker();
};

template <class T, bool t_doundo> void CWatchedValue<T, t_doundo>::removeListeners()
    {
     #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (c->hasListener(this)) {
          c->deleteWatch(this);
       }
       #endif
    }

template <class T, bool t_doundo> void CWatchedValue<T, t_doundo>::addChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata)
    {
        CUndoControl* c = CUndoControl::getInstance();
        c->addChangeListener(this, listener, changedata);
       
        
    }
template <class T, bool t_doundo> void CWatchedValue<T, t_doundo>::removeChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata)
    {
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
    }/*
template <class T, bool t_doundo> void CWatchedValue<T, t_doundo>::removeChangeListenerDeleteOK(CUndoChangeListener* listener, ReferenceUtiOB* changedata)
    {
      //Oh je, eventuell ist this hier weg!
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
    }*/
template <class T, bool t_doundo> void CWatchedValue<T, t_doundo>::set(const T& a) 
    {
       if (m_initialised && (m_data == a)) {return; }m_initialised =true;
       #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (t_doundo) {
       #else
       if (t_doundo) {
          CUndoControl* c = CUndoControl::getInstance();
       #endif
          if (!c->alreadyInThisSnapShot(this)) {
              c->addUndoCommand(new CWatchedUndo<T>(this, m_data));
          };
       };
       #ifdef UNDO_USECHANGE
       if (c->hasListener(this)) {
          c->preChange(this);
          m_data = a;
          c->postChange(this);
       } else {
       #endif
          m_data = a;
       #ifdef UNDO_USECHANGE
       }
       #endif
    }
    
    
template <class T, bool t_doundo> CWatchedValue<T, t_doundo>::~CWatchedValue()
    {
    #ifdef UNDO_USECHANGE
   // if (c->hasListener(this)) {
     // CUndoControl* c = CUndoControl::getInstance();
///TODO: Checke was dieses Assert soll
      //    utiassert(c->hasListener(this) == false);
      removeListeners();
       #endif
    }
    
 template <class T, bool t_doundo> void CWatchedReference<T, t_doundo>::removeListeners()
    {
     #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (c->hasListener(this)) {
          c->deleteWatch(this);
       }
       #endif
    }

template <class T, bool t_doundo> void CWatchedReference<T, t_doundo>::addChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata)
    {
        CUndoControl* c = CUndoControl::getInstance();
        c->addChangeListener(this, listener, changedata);
       
        
    }
template <class T, bool t_doundo> void CWatchedReference<T, t_doundo>::removeChangeListener(CUndoChangeListener* listener, ReferenceUtiOB* changedata)
    {
        CUndoControl* c = CUndoControl::getInstance();
        c->removeChangeListener(this, listener, changedata);
    }
template <class T, bool t_doundo> void CWatchedReference<T, t_doundo>::set(T* a) 
    {
       if (m_data == a) return;
       #ifdef UNDO_USECHANGE
       CUndoControl* c = CUndoControl::getInstance();
       if (t_doundo) {
       #else
       if (t_doundo) {
          CUndoControl* c = CUndoControl::getInstance();
       #endif
          //Immer hinzufügen, sonst wird Objekt gelöscht, dass Undo informationen enthält
          //if (!c->alreadyInThisSnapShot(this)) {
              c->addUndoCommand(new CWatchedReferenceUndo<T>(this, m_data.get()));
         //}
       };
       #ifdef UNDO_USECHANGE
       if (c->hasListener(this)) {
          c->preChange(this);
          m_data = a;
          c->postChange(this);
       } else {
       #endif
          m_data = a;
       #ifdef UNDO_USECHANGE
       }
       #endif
    }

#endif
