/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __MATH_CURVES__
#define __MATH_CURVES__

#include "Uti3DCore.h"
#include "Uti3DMath.h"
#include "utiarray.h"



template <class T> inline  T calculateBezier4(const T& p1, const T&p2, const T&p3, const T&p4, float u)
{
     float u1 = 1.0f-u;
     float b1 = u1*u1*u1;
     float b2 = 3*u*u1*u1;
     float b3 = 3*u*u*u1;
     float b4 = u*u*u;
     return p1*b1+p2*b2+p3*b3+p4*b4;
};


template <class T> inline  T calculateBezier4d(const T& p1, const T&p2, const T&p3, const T&p4, double u)
{
     double u1 = 1.0-u;
     double b1 = u1*u1*u1;
     double b2 = 3*u*u1*u1;
     double b3 = 3*u*u*u1;
     double b4 = u*u*u;
     return p1*b1+p2*b2+p3*b3+p4*b4;
};



template <class T> inline  T calculateBezier3(const T& p1, const T&p2, const T&p3, float u)
{
     float u1 = 1.0f-u;
     float b1 = u1*u1;
     float b2 = 2*u*u1;
     float b3 = u*u;
     return p1*b1+p2*b2+p3*b3;
};

template <class T> inline  T interpolate(const T& p1, const T&p2, float u)
{
     float u1 = 1.0f-u;
     float b1 = u1;
     float b2 = u;
     return p1*b1+p2*b2;
};

template<class T> struct SBezier
{
    StaticArray<T,4> m_points;
    void setBezier(const T& p1, const T&p2, const T&p3, const T&p4)
    {
        m_points[0] = p1;
        m_points[1] = p2;
        m_points[2] = p3;
        m_points[3] = p4;
    }
    void setLine(const T& p1, const T&p2)
    {
        m_points[0] = p1;
        m_points[1] = interpolate(p1, p2, 0.333333);
        m_points[2] = interpolate(p1, p2, 0.666666);
        m_points[3] = p2;
    }
    T eval(double d) const
    {
        return calculateBezier4d(m_points[0], m_points[1], m_points[2], m_points[3], d);
    }
    double length() const
    {
        T t1 = eval(0);
        double l=0.0;
        for (int i = 1; i <= 10; i++) {
            T tnext = eval(float(i)/float(10));
            l += t1.distanceTo(tnext);
            t1 = tnext;
        }
        return l;
    }
    void divideBezier(SBezier& bez1, SBezier& bez2, double t)
    {
        StaticArray2D<T, 4,4> b;
        for (int i = 0; i < 4; i++) {
            b(0, i) = m_points[i];

        }
        for (int i = 0; i < 3; i++) {
            b(1, i) = Interpolatef(b(0, i), b(0, i+1), t);

        }
        for (int i = 0; i < 2; i++) {
            b(2, i) = Interpolatef(b(1, i), b(1, i+1), t);

        }

        for (int i = 0; i < 1; i++) {
            b(3, i) = Interpolatef(b(2, i), b(2, i+1), t);

        }
        bez1.m_points[0] = b(0,0);
        bez1.m_points[1] = b(1,0);
        bez1.m_points[2] = b(2,0);
        bez1.m_points[3] = b(3,0);


        bez2.m_points[0] = b(3,0);
        bez2.m_points[1] = b(2,1);
        bez2.m_points[2] = b(1,2);
        bez2.m_points[3] = b(0,3);

    }
};

typedef SBezier<double> SBezier1;

typedef SBezier<CVector2> SBezier2;
typedef SBezier<CVector3> SBezier3;
typedef SBezier<CVector4> SBezier4;


void calculateBezierCurve(const CVector2& p1, const CVector2& p2, const CVector2& p3, const CVector2& p4, ArrayAccess< CVector2 >& result);
void calculateBezierCurveAdaptiv(const CVector2& p1, const CVector2& p2, const CVector2& p3, const CVector2& p4, float pixelsize, float intolerance, ArrayAccess< CVector2 >& result);



#endif
