/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __EQUATIONS__
#define __EQUATIONS__
#include "matrices.h"
#include <cmath>

const double BISECTION_LIMIT=0.01f;
const double SEKANTE_LIMIT=0.01f;
const int MAXSTEPS= 64;

template<class T, class Func> T bisection(const T& a, const T& b, const Func& f, bool& hassolution, const T& limit=T(BISECTION_LIMIT))
{
    T mya = a;
    T myb = b;
  T fa = f(a);
  T fb = f(b);
  hassolution = true;
  if (std::abs(fa) < limit) {
     // UMat<T> ret(a);
      return a;
   }
  if (std::abs(fb) < limit) {
    //  UMat<T> ret(b);
      return b;
   }
  if (fa*fb >= T(0)) {
    //  UMat<T> ret(0,0);
      hassolution = false;
      return a;
   }
   int steps = 0;
  while (std::abs(mya-myb) > limit && steps < MAXSTEPS) {
    T mid = (mya+myb)/T(2);
    T fmid = f(mid);
    if (fa*fmid< T(0)) {
     // b = mid;
      myb = mid;
    } else {
     // a = mid;
      mya = mid;
    };
    steps++;
  };
  if (steps >= MAXSTEPS) {
    // UMat<T> ret(0,0);
      hassolution = false;
      return mya;
  };
  {
   //   UMat<T> ret(a);
      return (mya+myb)/2;
  }
};


template<class T, class Func> UMat<T> regularFalsi(const T& a, const T& b, const Func& f, const T& limit=T(BISECTION_LIMIT))
{
  T fa = f(a);
  T fb = f(b);
  if (std::abs(fa) < limit) {
      UMat<T> ret(a);
      return ret;
   }
  if (std::abs(fb) < limit) {
      UMat<T> ret(b);
      return ret;
   }
  if (fa*fb >= T(0)) {
      UMat<T> ret(0,0);
      return ret;
   }
   int steps = 0;
  while (std::abs(fa-fb) > limit && steps < MAXSTEPS) {
    T mid = (a*fb-b*fa)/(fb-fa);
    T fmid = f(mid);
    if (fa*fmid< T(0)) {
      b = mid;
      fb = fmid;
    } else {
      a = mid;
      fa = fmid;
    };
    steps++;
  };
  if (steps >= MAXSTEPS) {
     UMat<T> ret(0,0);
      return ret;
  };
  {
      UMat<T> ret(a);
      return ret;
  }
};


template<class T, class Func> UMat<T> sekante(const T& a, const T& b, const Func& f, const T& limit=T(SEKANTE_LIMIT))
{
  T fa = f(a);
  T fb = f(b);
  if (std::abs(fa) < limit) {
      UMat<T> ret(a);
      return ret;
   }
  if (std::abs(fb) < limit) {
      UMat<T> ret(b);
      return ret;
   }
   if (fa*fb >= T(0)) {
     // UMat<T> ret(0,0);
     // return ret;
     //
   }
   int steps = 0;
   T lastxpositive;
   T lastypositive;
   bool haspositive;
   
   T lastxnegative;
   T lastynegative;
   bool hasnegative;
    bool nextvaluepositiv =false;
   if (fa < T(0)) {
     lastxnegative = a;
     lastynegative = fa;
      hasnegative = true;
       nextvaluepositiv =false;
   } else {
     lastxpositive = a;
     lastypositive = fa;
     haspositive = true;
        nextvaluepositiv =true;
   };
   
   if (fb < T(0)) {
     lastxnegative = b;
     lastynegative = fb;
     hasnegative = true;
     if (haspositive == false) {
       lastxpositive = a;
       lastypositive = fa;
          nextvaluepositiv =true;
     };
   } else {
     lastxpositive = b;
     lastypositive = fb;
     haspositive = true;
     if (hasnegative == false) {
       lastxnegative = a;
       lastynegative = fa;
          nextvaluepositiv =false;
     };
   };
  
  while (std::abs(lastypositive-lastynegative) > limit && steps < MAXSTEPS) {
    T mid;
    T fmid;
    if (nextvaluepositiv) {
      mid = lastxpositive - lastypositive* (lastxpositive- lastxnegative)/(lastypositive - lastynegative);
    } else {
      mid = lastxnegative - lastynegative* (lastxnegative- lastxpositive)/(lastynegative - lastypositive);
    };
    fmid = f(mid);
    if (std::abs(fmid) < limit) {
      UMat<T> ret(mid);
      return ret;
    };
    if (fmid < T(0)) {
      lastxnegative = mid;
      lastynegative = fmid;
     
      nextvaluepositiv = false;
    } else {
      lastxpositive = mid;
      lastypositive = fmid;
     
      nextvaluepositiv = true;
    };
    steps++;
  };
  {
     UMat<T> ret(0,0);
      return ret;
  };
  
};

template<class T, class Func> T newton(const T& xstart,  const Func& f, const Func& derv, const T& limit=T(SEKANTE_LIMIT))
{
  T fx = f(xstart);
  T fderv = derv(xstart);
  if (std::abs(fx) < limit) {
     
      return xstart;
   }
   T x = xstart;
   int steps = 0;
   while (std::abs(fx) > limit && std::abs(fderv) > limit && steps < MAXSTEPS) {
    T next = x - fx/fderv;
    x = next;
    steps++;
  };
  return x;
}



#endif
