#ifndef __FIXEDPOINT__
#define __FIXEDPOINT__

#include <limits>
#include <string>
#include <math.h>

#include "gdef.h"

template <typename T2> struct enhancedtype {typedef T2 type;};
    template <> struct enhancedtype<signed char> { typedef signed short type;}; 
    template <> struct enhancedtype<unsigned char> { typedef unsigned short type;}; 
    template <> struct enhancedtype<signed short> { typedef signed int type;}; 
    template <> struct enhancedtype<unsigned short> { typedef unsigned int type;}; 
    template <> struct enhancedtype<signed int> { typedef UTI_INT64 type;}; 
    template <> struct enhancedtype<unsigned int> { typedef UTI_UINT64 type;}; 
    template <> struct enhancedtype<UTI_INT64> { typedef UTI_INT64 type;}; 
template<typename T, unsigned char INTEGERPART, unsigned char FRACTIONALPART= std::numeric_limits<T>::digits -INTEGERPART>
struct CBaseFixedPoint
{
  protected:
    typedef CBaseFixedPoint<T, INTEGERPART, FRACTIONALPART> NumberType;
    typedef typename enhancedtype<T>::type EnhancedNumberType;
    T m_value;
    
  public:
    enum {INTEGERMUL=(1 << FRACTIONALPART), INTEGERAND=(1 << FRACTIONALPART)-1};
    enum {INT_PART = INTEGERPART, FRAC_PART = FRACTIONALPART};
    inline CBaseFixedPoint() {};
    inline CBaseFixedPoint(const CBaseFixedPoint& f) {m_value = f.m_value;};
    inline CBaseFixedPoint(int f) { m_value = ((T)f) << FRACTIONALPART;};
    inline CBaseFixedPoint(long f) { m_value = ((T)f) << FRACTIONALPART;};
    inline CBaseFixedPoint(double f) { m_value = (T)(f*(double)INTEGERMUL);};
    inline CBaseFixedPoint(float f) { m_value = (T)(f*(float)INTEGERMUL);};
    template<unsigned char I, unsigned char F> inline  CBaseFixedPoint(const CBaseFixedPoint<T, I, F> & t) {
      *this = t;
    };
    inline T fixedvalue() const { return m_value;};
    inline operator float() const {return (float) m_value /  INTEGERMUL;};
    inline operator double() const {return (double) m_value /  INTEGERMUL;};
    inline operator int() const {return (int)( m_value >> FRACTIONALPART);};
    inline operator long() const {return (long)( m_value >> FRACTIONALPART);};
    inline T intpart() const {return m_value >> FRACTIONALPART;};
    inline T fractionalpart() const {return m_value & INTEGERAND;};
    inline void swap(NumberType & other)
    {
       std::swap(m_value, other.m_value);
    };
    inline NumberType& operator=(const NumberType & t)
    {
      m_value = t.m_value;
      return *this;
    };
    template<unsigned char I, unsigned char F> inline NumberType& operator=(const CBaseFixedPoint<T, I, F> & t)
    {
      if (I > INTEGERPART) {
	m_value = m_value << (I-INTEGERPART);
      } else {
	m_value = m_value >> (INTEGERPART-I);
      };
      return *this;
    };
    inline bool operator<(const NumberType& t) const {return m_value < t.m_value;};
    inline bool operator==(const NumberType& t) const {return m_value == t.m_value;};
    inline bool operator>(const NumberType& t) const {return m_value > t.m_value;};
    inline bool operator<=(const NumberType& t) const {return m_value <= t.m_value;};
    inline bool operator>=(const NumberType& t) const {return m_value >= t.m_value;};
    inline bool operator!=(const NumberType& t) const {return m_value != t.m_value;};
    inline operator bool() const { return (bool) m_value;};
    inline bool operator!() const {return m_value == 0;};
    inline NumberType& operator-() const {NumberType result; result.m_value = -m_value; return result;};
    inline NumberType& operator++() {m_value += INTEGERMUL; return *this;};
    inline NumberType& operator--() {m_value -= INTEGERMUL; return *this;};
    inline NumberType operator++(int) const {NumberType result; result.m_value = m_value; ++result; return result;};
    inline NumberType operator--(int) const {NumberType result; result.m_value = m_value; --result; return result;};
    inline NumberType& operator+=(const NumberType& t) { m_value += t.m_value; return *this;};
    inline NumberType& operator-=(const NumberType& t) {m_value -= t.m_value; return *this;};
    inline NumberType operator+(const NumberType& t) const {NumberType result(*this);  result.m_value += t.m_value;return result;};
    inline NumberType operator-(const NumberType& t) const {NumberType result(*this); result.m_value -= t.m_value; return result;};
    inline NumberType& operator*=(const NumberType& t) 
    { 
      
      m_value = (((EnhancedNumberType)m_value) * ((EnhancedNumberType)t.m_value)) >> FRACTIONALPART; 
      return *this;
    };
    
    inline NumberType& operator/=(const NumberType& t)
    {
      m_value = (((EnhancedNumberType)m_value) << FRACTIONALPART) / ((EnhancedNumberType)t.m_value); 
      return *this;
    };
    
    inline NumberType operator*(const NumberType& t) const {NumberType result(*this);  result *= t;return result;};
    inline NumberType operator/(const NumberType& t) const {NumberType result(*this); result /= t; return result;};
    
    inline NumberType& operator*=(int t) 
    { 
      m_value *= t; return *this;
    };
    
    inline NumberType& operator/=(int t)
    {
      m_value /= t; 
      return *this;
    };
    
    inline NumberType operator*(int t) const {NumberType result(*this);  result *= t;return result;};
    inline NumberType operator/(int t) const {NumberType result(*this); result /= t; return result;};
    
    inline NumberType& operator*=(long t) 
    { 
      m_value *= t; return *this;
    };
    
    inline NumberType& operator/=(long t)
    {
      m_value /= t; 
      return *this;
    };
    
    inline NumberType operator*(long t) const {NumberType result(*this);  result *= t;return result;};
    inline NumberType operator/(long t) const {NumberType result(*this); result /= t; return result;};
    inline NumberType& operator>>=(size_t shift) {m_value >>= shift; return *this;};
    inline NumberType& operator<<=(size_t shift) {m_value <<= shift; return *this;};
    inline NumberType operator>>(size_t shift) const {NumberType result(*this);  result >>= shift;return result;};
    inline NumberType operator<<(size_t shift) const {NumberType result(*this); result <<=shift; return result;};
    inline NumberType fabs() {return m_value < 0 ? -(*this) : (*this);};
    inline friend NumberType fabs(NumberType t) {return t.fabs();};
    
   inline friend NumberType sin(NumberType x)
   {
     return sin((double)x);
   };
   inline friend NumberType cos(NumberType x)
   {
     return cos((double)x);
   };
   inline friend NumberType tan(NumberType x)
   {
     return tan((double)x);
   };
    inline friend NumberType sqrt(NumberType x)
   {
     return sqrt((double)x);
   };
   inline friend NumberType exp(NumberType x)
   {
     return exp((double)x);
   };
};

template<typename S, typename T, unsigned char INTEGERPART, unsigned char FRACTIONALPART>
   S& operator >>(S & s, CBaseFixedPoint<T, INTEGERPART, FRACTIONALPART> & value)
   {
     double v = 0.0;
     s >> v;
     if (s) value =  v;
     return s;
   };
   
template<typename S, typename T, unsigned char INTEGERPART, unsigned char FRACTIONALPART>
   S& operator <<(S & s, const CBaseFixedPoint<T, INTEGERPART, FRACTIONALPART> & value)
   {
     double v = value;
     s << v;
    
     return s;
   };


typedef CBaseFixedPoint<int, 16> Fixed16;
typedef CBaseFixedPoint<UTI_INT64, 32> Fixed32;

#endif