/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __INTERPOLATIONS__
#define __INTERPOLATIONS__
#include "Uti3DMath.h"

template<class T, int t_numerator_speed=1, int t_denominator_speed=1> class STimeInterpolation
{
protected:
   float m_speed;
   T m_value;
   T m_target;
   bool m_hasvalue;

public:
   typedef T value_type;
   STimeInterpolation(float speed=1.0f, T value=T()) {
       m_hasvalue = 0;
       m_speed = (float)t_numerator_speed / (float)t_denominator_speed;
       m_value = value;
       m_target = value;
   }
   inline void setSpeed(float s)
   {
       m_speed = s;
   }
   inline void setValue(T v)
   {
       m_value = v;
       m_target = v;
       m_hasvalue = true;
   }
   inline void setValueInterpolate(T v)
   {
       m_target = v;
       if (m_hasvalue) {

       } else {
           setValue(v);
       }
   }
   inline void timeStep(TIME t)
   {
       float bt = uclamp(t*m_speed,0.0f,1.0f);
       m_value = m_value*(1.0f-bt)+m_target*bt;
   }
   inline T getValue() const
   {
       return m_value;
   }
};


template<class T, int t_numerator_speed=1, int t_denominator_speed=1> class STimeLinear
{
protected:
   float m_speed;
   T m_value;
   T m_target;
   bool m_hasvalue;
public:
   typedef T value_type;
   STimeLinear(float speed=1.0f, T value=T()) {
       m_hasvalue = 0;
       m_speed = (float)t_numerator_speed / (float)t_denominator_speed;
       m_value = value;
       m_target = value;
   }
   inline void setSpeed(float s)
   {
       m_speed = s;
   }
   void setValue(T v)
   {
       m_value = v;
       m_target = v;
       m_hasvalue = true;
   }
   void setValueInterpolate(T v)
   {
       m_target = v;
       if (m_hasvalue) {

       } else {
           setValue(v);
       }
   }
   void timeStep(TIME t)
   {
       if (m_target < m_value) {
           m_value -= m_speed*t;
           if (m_target > m_value) {
               m_value = m_target;
           }
           return;
       }
       if (m_target > m_value) {
           m_value += m_speed*t;
           if (m_target < m_value) {
               m_value = m_target;
           }
           return;
       }
   }

   inline T getValue() const
   {
       return m_value;
   }
};

template<class Interpolation, int count> class STimeArray
{
protected:
    Interpolation m_array[count];

public:
    typedef typename Interpolation::value_type value_type;
    enum {COUNT=count};
    STimeArray()
    {

    }
    void setValue(int index, value_type t)
    {
        assert(index >= 0 && index < count);
        m_array[index].setValue(t);
    }
    void setValueInterpolate(int index, value_type t)
    {
        assert(index >= 0 && index < count);
        m_array[index].setValueInterpolate(t);
    }
    inline value_type getValue(int index) const
    {
        assert(index >= 0 && index < count);
        return m_array[index].getValue();

    }
    void setSpeed(float s)
       {
        for (int i = 0; i < count; i++) {
            m_array[i].setSpeed(s);
        }
       }
    void timeStep(TIME t)
    {
        for (int i = 0; i < count; i++) {
            m_array[i].timeStep(t);
        }
    }
};

template<class T,class time_array, int start = 0, int count=time_array::COUNT> class STimeArray_Access
{
protected:
    time_array* m_timearray;
public:
    STimeArray_Access()
    {
        m_timearray = NULL;
    }
    STimeArray_Access(time_array& t)
    {
        m_timearray =&t;
    }
    inline void setValue(const T& v)
    {
        assert(m_timearray != 0);
        for (int i = 0; i < count;i++ ) {
            m_timearray->setValue(i, v[0]);
        }
    }
    inline void setValueInterpolate(const T& v)
    {
        assert(m_timearray != 0);
        for (int i = 0; i < count;i++ ) {
            m_timearray->setValueInterpolate(i, v[0]);
        }
    }
    inline T getValue()
    {
        T v;
        assert(m_timearray != 0);
        for (int i = 0; i < count;i++ ) {
            v[i] = m_timearray->getValue(i);
        }
        return v;
    }
};

template<class T,class time_array, int start = 0, int count=time_array::COUNT> class STimeArray_Container
{
protected:
    time_array m_array;
    STimeArray_Access<T, time_array, start, count> m_access;
public:
    STimeArray_Container() :m_array(), m_access(m_array)
    {

    }
    inline void setValue(const T& v)
    {
        m_access.setValue(v);
    }
    inline void setValueInterpolate(const T& v)
    {
        m_access.setValueInterpolate(v);
    }
    inline T getValue()
    {
        return m_access.getValue();
    }
    void setSpeed(float s)
       {
        m_array.setSpeed(s);
       }
    void timeStep(TIME t)
    {
        m_array.timeStep(t);
    }

};

template<class Interpolation> struct STime_CVector2 : public STimeArray_Container<CVector2, STimeArray<Interpolation, 2> > {};
template<class Interpolation> struct STime_CVector3 : public STimeArray_Container<CVector3, STimeArray<Interpolation, 3> > {};
template<class Interpolation> struct STime_CVector4 : public STimeArray_Container<CVector4, STimeArray<Interpolation, 4> > {};

void createBlurArray(Array<float>& blur, float sigma, int &center);
void createBlurStepArray(Array<float>& blur, float sigma,int k, int &center);

#endif
