#ifndef NUMERIC_H
#define NUMERIC_H

#include "matrices.h"

template <class T, class Func, class DFunc> UMat<T> gauss_newton(Func f, Func df, const UMat<T>& x0, int maxit, T tol)
{
   int k = 0;
   UMat<T> x = x0;
   UMat<T> s, s2, dfmat;
   dfmat = df(x0);
   s = -(dfmat.transpose()*dfmat);
   s2 = dfmat.transpose()*f(x0);


   s = solveSystem(s, s2);

   while (norm(s) > tol && k < maxit) {
       k = k + 1;
       x = x + s;

       dfmat = df(x);
       s = -(dfmat.transpose()*dfmat);
       s2 = dfmat.transpose()*f(x);
       s = solveSystem(s, s2);
   }
   return x;
}

template <class T, class Func, class DFunc>  std::pair<UMat<T>, T> levenberg_marquardt_korrektur(Func f, Func df, const UMat<T>& x, T mu, T beta0, T beta1)
{

    int n = x.sizey();

    UMat<T> s, s2, dfmat, fmat, fmats,f2, f3;
    dfmat = df(x);
    fmat = f(x);
    s = -(dfmat.transpose()*dfmat+pow(mu, 2)*UMat<T>::identity(n));
    s2 = dfmat.transpose()*fmat;
    s = solveSystem(s, s2);

    fmat = f(x+s);
    f2 = fmat.transpose()*fmat;
    f3 = (fmat+dfmat*s);
    T eps_mu = solveSystem(f2-fmats.transpose()*fmats, f2-f3.transpose()*f3);
    if (eps_mu <= beta0) {
        return levenberg_marquardt_korrektur(f, df, x, 2*mu, beta0, beta1);
    }else if (eps_mu >= beta1) {
        mu = mu/2;
    }
    return std::pair<UMat<T>, T>(fmat, mu);
}

template <class T, class Func, class DFunc> UMat<T> levenberg_marquardt(Func f, Func df, const UMat<T>& x0, T mu0, T beta0, T beta1, int maxit, T tol)
{
    int n = x0.sizey();
    int k = 0;
    T mu = mu0;
    UMat<T> x = x0;

    UMat<T> s, s2, dfmat;
    dfmat = df(x0);
    s = -(dfmat.transpose()*dfmat+pow(mu, 2)*UMat<T>::identity(n));
    s2 = dfmat.transpose()*f(x0);

    s = solveSystem(s, s2);

    while (norm(s) > tol && k < maxit) {
        std::pair<UMat<T>, T> res = levenberg_marquardt_korrektur(f, df, x, mu, beta0, beta1);
        k = k + 1;
        x = x + res.first;


    }
    return x;
}

template <class T> std::pair<UMat<T>, T> potenzMethode(const UMat<T>& A, const UMat<T>& x, int steps)
{
//assert v_1^T*x != 0 and norm(x) == 1
    UMat<T> a, xcurrent;
    xcurrent = x;
    float p;
    for (int i = 0; i < steps; i++) {
        a = A.multiply(xcurrent);
        p = xcurrent.transpose().multiply(a);
        xcurrent = a / norm(a);
    }
    return std::make_pair(xcurrent, p);
}

template <class T> std::pair<UMat<T>, T> wielandIteration(const UMat<T>& A, const UMat<T>& x, T mu, int steps)
{
//assert v_1^T*x != 0 and norm(x) == 1

    CULUDecomposition<T> decompose(A-mu*UMat<T>::identity(A.sizex()));

    UMat<T> a, xcurrent;
    xcurrent = x;
    float p;
    for (int i = 0; i < steps; i++) {
        a = decompose.solveSystem(xcurrent);
        p = xcurrent.transpose().multiply(a);
        xcurrent = a / norm(a);
    }
    return std::make_pair(xcurrent, p);
}

template<class T> UMat<T> qrBasisIteration(const UMat<T> A, int steps)
{
    UMat<T> AC, Q, R;
    AC = A;
    for (int i = 0; i < steps; i++) {
        qrDecomposition(AC, Q, R);
        AC = R*Q;
    }
    return AC;
}


//Rasterization

struct SRasterScreen {
    float m_width;
    float m_height;
    int m_widthi;
    int m_heighti;
    float m_halfwidth;
    float m_halfheight;
    int m_halfwidthi;
    int m_halfheighti;
    float m_focal;
    void setup(float w, float h, float openangle);
    inline float projX(float px, float pz) const {return m_focal*px/pz;}
    inline float projY(float py, float pz) const {return m_focal*py/pz;}
    inline float projX(const CVector3f& p) const {return m_focal*p.x/p.z;}
    inline float projY(const CVector3f& p) const {return m_focal*p.y/p.z;}
};


typedef std::pair<float, float> SRasterInterval;

struct SRaster3DLine {
    CVector3f m_x1;
    CVector3f m_x2;
    CVector3f m_d;

    inline SRaster3DLine(const CVector3f &x1, const CVector3f& x2)
    {
        m_x1 = x1;
        m_x2 = x2;
        m_d = x2-x1;
    }

    inline float getT_by_SX(const SRasterScreen& screen, float sx) //sx ist in der Mitte 0
    {
        float denominator = (sx*m_d.z-screen.m_focal*m_d.x);
        if (denominator < 0.0001f || denominator > -0.0001f) return 0.0f;
        return (screen.m_focal* m_x1.x - sx*m_x1.z) / denominator;
    }

    inline SRasterInterval getPaintInterval(const SRasterScreen& screen)
    {
        float t0 = getT_by_SX(screen, -screen.m_halfwidth);
        float t1 = getT_by_SX(screen, screen.m_halfwidth);
        if (t0 < t1) {
            if (t0 > 1.0f || t1 < 0.0f) return std::make_pair(1000,-1000);
            float minvalue; float maxvalue;
            if (t0 > 0.0f) minvalue = -screen.m_halfwidth;
            if (t0 < 0.0f) minvalue = screen.projX(m_x1);

            if (t1 > 1.0f) maxvalue = screen.projX(m_x2);
            if (t1 > 0.0f) maxvalue = screen.m_halfwidth;

            return std::make_pair(minvalue, maxvalue);

        } else {
            if (t0 > 0.0f || t1 > 1.0f) return std::make_pair(1000,-1000);
            float minvalue; float maxvalue;
            if (t0 > 1.0f) minvalue = screen.projX(m_x2);
            if (t0 > 0.0f) minvalue = -screen.m_halfwidth;

            if (t1 < 0.0f) maxvalue = screen.projX(m_x2);
            if (t1 < 1.0f) maxvalue = screen.m_halfwidth;
            return std::make_pair(minvalue, maxvalue);
        }
    }
};

inline SRasterInterval union_interval(const SRasterInterval& i1, const SRasterInterval& i2)
{
    SRasterInterval res;
    res.first = std::min(i1.first, i2.first);
    res.second = std::max(i1.second, i2.second);

    return res;
}

struct SSpanCollection
{
    bool m_valid;
    int m_minx;
    int m_maxy;
    SRaster3DLine m_top;
    SRaster3DLine m_bottom;
};

struct SRasterTriangle {
    StaticArray<CVector3f, 3> m_vertices;
    CVector3f m_normal;

    void createSpawnCollection(const SRasterScreen& screen, SSpanCollection& s1, SSpanCollection &s2);
};



#endif // NUMERIC_H
