/***************************************************************************
 *   Copyright (C) 2007 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef NUMERIC_H
#define NUMERIC_H

#include "matrices.h"
#include "utiadvarray.h"

struct CollisionContactPoint
{
    CVector3 m_worldposition;
    double m_penetrationdepth;
    CVector3 m_penetrationnormal;
};

bool collideBoxBoxAdvanced(const COBB& box1, const COBB& box2, Array<CollisionContactPoint>& rescontactpoints, int& rescount, double &maxpenetration);

template<class T, class List=DynArray1D<T> > class CStatisticAnalyser : public ReferenceUtiOB
{
public:
    List m_data;
    T median() const
    {

        int amount = m_data.begin()-m_data.end();
        if (amount == 0) return T();
        if ((amount % 2) == 0) {
            int pos = amount / 2;
            if (pos >= amount) return T();
            if (pos -1 < 0) return T();

            return (*(m_data.begin()+(pos-1))+*(m_data.begin()+(pos))) / T(2);

        } else {
            int pos = amount / 2;
            if (pos >= amount) return T();
            return *(m_data.begin()+pos);
        }
    }
    T mean() const
    {
        int amount = 0;
        T res = T();
        for (auto el: m_data) {
            res += el;
            amount++;
        }
        if (amount == 0) return T();
        res /= T(amount);
        return res;

    }
    T variance() const
    {
        int amount = 0;
        T m1 = T();
        T m2 = T();
        for (auto el: m_data) {
            m1 += el;
            m2 += el*el;
            amount++;
        }
        if (amount == 0) return T();
        m1 /= T(amount);
        m2 /= T(amount);
        return m2-m1*m1;
    }
    T stdDerivation() const
    {
        return sqrt(variance());
    }
    T powerMean(double p) const
    {
        int amount = 0;
        T res = T();
        for (auto el: m_data) {
            res += pow(el, p);
            amount++;
        }
        if (amount == 0) return T();
        res /= T(amount);
        res = pow(res, 1.0f/p);
        return res;
    }
    T harmonicMean() const
    {
        int amount = 0;
        T res = T();
        for (auto el: m_data) {
            res += T(1)/el;
            amount++;
        }
        if (amount == 0) return T();
        res = T(amount) / res;

        return res;
    }
    T geometricMean() const
    {
        int amount = 0;
        T res = T();
        for (auto el: m_data) {
            res *= el;
            amount++;
        }
        if (amount == 0) return T();
        res = pow(res, 1.0/double(amount));

        return res;
    }
    T maximum() const
    {
        if (m_data.begin() == m_data.end()) return 0.0;
        T mres = *m_data.begin();
        for (auto el: m_data) {
            if (el > mres) {
                mres = el;
            }
        }


        return mres;
    }
    T minimum() const
    {
        if (m_data.begin() == m_data.end()) return 0.0;
        T mres = *m_data.begin();
        for (auto el: m_data) {
            if (el < mres) {
                mres = el;
            }
        }


        return mres;
    }
    int maximumIndex() const
    {
        if (m_data.begin() == m_data.end()) return 0.0;
        T mres = *m_data.begin();
        int resindex = 0;
        int currentindex=0;
        for (auto el: m_data) {
            if (el > mres) {
                mres = el;
                resindex = currentindex;
            }
            currentindex++;
        }


        return resindex;
    }
    int minimumIndex() const
    {
        if (m_data.begin() == m_data.end()) return 0.0;
        T mres = *m_data.begin();
        int resindex = 0;
        int currentindex=0;
        for (auto el: m_data) {
            if (el < mres) {
                mres = el;
                resindex = currentindex;
            }
            currentindex++;
        }

        return resindex;
    }
};

template <class T, class Func, class DFunc> UMat<T> gauss_newton(Func f, Func df, const UMat<T>& x0, int maxit, T tol)
{
   int k = 0;
   UMat<T> x = x0;
   UMat<T> s, s2, dfmat;
   dfmat = df(x0);
   s = -(dfmat.transpose()*dfmat);
   s2 = dfmat.transpose()*f(x0);


   s = solveSystem(s, s2);

   while (norm(s) > tol && k < maxit) {
       k = k + 1;
       x = x + s;

       dfmat = df(x);
       s = -(dfmat.transpose()*dfmat);
       s2 = dfmat.transpose()*f(x);
       s = solveSystem(s, s2);
   }
   return x;
}

template <class T, class Func, class DFunc>  std::pair<UMat<T>, T> levenberg_marquardt_korrektur(Func f, Func df, const UMat<T>& x, T mu, T beta0, T beta1)
{

    int n = x.sizey();

    UMat<T> s, s2, dfmat, fmat, fmats,f2, f3;
    dfmat = df(x);
    fmat = f(x);
    s = -(dfmat.transpose()*dfmat+pow(mu, 2)*UMat<T>::identity(n));
    s2 = dfmat.transpose()*fmat;
    s = solveSystem(s, s2);

    fmat = f(x+s);
    f2 = fmat.transpose()*fmat;
    f3 = (fmat+dfmat*s);
    T eps_mu = solveSystem(f2-fmats.transpose()*fmats, f2-f3.transpose()*f3);
    if (eps_mu <= beta0) {
        return levenberg_marquardt_korrektur(f, df, x, 2*mu, beta0, beta1);
    }else if (eps_mu >= beta1) {
        mu = mu/2;
    }
    return std::pair<UMat<T>, T>(fmat, mu);
}

template <class T, class Func, class DFunc> UMat<T> levenberg_marquardt(Func f, Func df, const UMat<T>& x0, T mu0, T beta0, T beta1, int maxit, T tol)
{
    int n = x0.sizey();
    int k = 0;
    T mu = mu0;
    UMat<T> x = x0;

    UMat<T> s, s2, dfmat;
    dfmat = df(x0);
    s = -(dfmat.transpose()*dfmat+pow(mu, 2)*UMat<T>::identity(n));
    s2 = dfmat.transpose()*f(x0);

    s = solveSystem(s, s2);

    while (norm(s) > tol && k < maxit) {
        std::pair<UMat<T>, T> res = levenberg_marquardt_korrektur(f, df, x, mu, beta0, beta1);
        k = k + 1;
        x = x + res.first;


    }
    return x;
}

template <class T> std::pair<UMat<T>, T> potenzMethode(const UMat<T>& A, const UMat<T>& x, int steps)
{
//assert v_1^T*x != 0 and norm(x) == 1
    UMat<T> a, xcurrent;
    xcurrent = x;
    float p;
    for (int i = 0; i < steps; i++) {
        a = A.multiply(xcurrent);
        p = xcurrent.transpose().multiply(a);
        xcurrent = a / norm(a);
    }
    return std::make_pair(xcurrent, p);
}

template <class T> std::pair<UMat<T>, T> wielandIteration(const UMat<T>& A, const UMat<T>& x, T mu, int steps)
{
//assert v_1^T*x != 0 and norm(x) == 1

    CULUDecomposition<T> decompose(A-mu*UMat<T>::identity(A.sizex()));

    UMat<T> a, xcurrent;
    xcurrent = x;
    float p;
    for (int i = 0; i < steps; i++) {
        a = decompose.solveSystem(xcurrent);
        p = xcurrent.transpose().multiply(a);
        xcurrent = a / norm(a);
    }
    return std::make_pair(xcurrent, p);
}

template<class T> UMat<T> qrBasisIteration(const UMat<T> A, int steps)
{
    UMat<T> AC, Q, R;
    AC = A;
    for (int i = 0; i < steps; i++) {
        qrDecomposition(AC, Q, R);
        AC = R*Q;
    }
    return AC;
}


//Rasterization

struct SRasterScreen {
    float m_width;
    float m_height;
    int m_widthi;
    int m_heighti;
    float m_halfwidth;
    float m_halfheight;
    int m_halfwidthi;
    int m_halfheighti;
    float m_focal;
    void setup(int w, int h, float openangle);
    inline float projX(float px, float pz) const {return m_focal*px/pz;}
    inline float projY(float py, float pz) const {return m_focal*py/pz;}
    inline float projX(const CVector3f& p) const {return m_focal*p.x/p.z;}
    inline float projY(const CVector3f& p) const {return m_focal*p.y/p.z;}
};


typedef std::pair<float, float> SRasterInterval;

struct SRaster3DLine {
    CVector3f m_x1;
    CVector3f m_x2;
    CVector3f m_d;

    inline SRaster3DLine(const CVector3f &x1, const CVector3f& x2)
    {
        m_x1 = x1;
        m_x2 = x2;
        m_d = x2-x1;
    }

    inline float getT_by_SX(const SRasterScreen& screen, float sx) //sx ist in der Mitte 0
    {
        float denominator = (sx*m_d.z-screen.m_focal*m_d.x);
        if (denominator < 0.0001f || denominator > -0.0001f) return 0.0f;
        return (screen.m_focal* m_x1.x - sx*m_x1.z) / denominator;
    }

    inline SRasterInterval getPaintInterval(const SRasterScreen& screen)
    {
        float t0 = getT_by_SX(screen, -screen.m_halfwidth);
        float t1 = getT_by_SX(screen, screen.m_halfwidth);
        if (t0 < t1) {
            if (t0 > 1.0f || t1 < 0.0f) return std::make_pair(1000,-1000);
            float minvalue; float maxvalue;
            if (t0 > 0.0f) minvalue = -screen.m_halfwidth;
            if (t0 < 0.0f) minvalue = screen.projX(m_x1);

            if (t1 > 1.0f) maxvalue = screen.projX(m_x2);
            if (t1 > 0.0f) maxvalue = screen.m_halfwidth;

            return std::make_pair(minvalue, maxvalue);

        } else {
            if (t0 > 0.0f || t1 > 1.0f) return std::make_pair(1000,-1000);
            float minvalue; float maxvalue;
            if (t0 > 1.0f) minvalue = screen.projX(m_x2);
            if (t0 > 0.0f) minvalue = -screen.m_halfwidth;

            if (t1 < 0.0f) maxvalue = screen.projX(m_x2);
            if (t1 < 1.0f) maxvalue = screen.m_halfwidth;
            return std::make_pair(minvalue, maxvalue);
        }
    }
};

inline SRasterInterval union_interval(const SRasterInterval& i1, const SRasterInterval& i2)
{
    SRasterInterval res;
    res.first = std::min(i1.first, i2.first);
    res.second = std::max(i1.second, i2.second);

    return res;
}

inline bool interval_valid(const SRasterInterval& v1)
{
   return v1.first <= v1.second;
}

inline float interval_length(const SRasterInterval& v1)
{
    assert(interval_valid(v1));
    return v1.second-v1.first;
}

class SSpanCollection;

struct SRasterTriangle {
    StaticArray<CVector3f, 3> m_vertices;
    CPlanef m_plane;

    void calculatePlane();
    void createSpawnCollection(const SRasterScreen& screen, SSpanCollection& s1, SSpanCollection &s2);
};

struct SSpanCollection
{
    bool m_valid;
    int m_min;
    int m_max;
    SRasterTriangle* m_triangle;
    SRaster3DLine m_top;
    SRaster3DLine m_bottom;
};

class SSpan{
public:
    int m_miny;
    int m_maxy;
};

// kdtree = new CKDTree(this, new CKDGroupAccess(this));
//kdtree->compile();

class CRaycastObject : public ReferenceUtiOB {
public:
};


template<class T> float area_triangle(const T& v1, const T & v2, const T & v3)
{
    float a = v1.distanceTo(v2);
    float b = v2.distanceTo(v3);
    float c = v3.distanceTo(v1);

    float s = (a+b+c)/2.0f;
    float area = sqrt(s*(s-a)*(s-b)*(s-c));
    return area;
}

inline CVector2 projectViaAxis(int majoraxis, const CVector3& pos)
{
    switch (majoraxis) {
    case 0: {return pos.yz();}; break;
    case 1: {return pos.xz();}; break;
    case 2: {return pos.xy();}; break;
    default:{assert(false);}; break;
    }
    assert(false);
    return pos.xy();
}

inline CVector2f projectViaAxis(int majoraxis, const CVector3f& pos)
{
    switch (majoraxis) {
    case 0: {return pos.yz();}; break;
    case 1: {return pos.xz();}; break;
    case 2: {return pos.xy();}; break;
    default:{assert(false);}; break;
    }
    assert(false);
    return pos.xy();
}



#endif // NUMERIC_H
