#include "numericalrecipes.h"

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#define NR_END 1
#define FREE_ARG char*

void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
        fprintf(stderr,"Numerical Recipes run-time error...\n");
        fprintf(stderr,"%s\n",error_text);
        fprintf(stderr,"...now exiting to system...\n");
        exit(1);
}

float *vector(long nl, long nh)
/* allocate a float vector with subscript range v[nl..nh] */
{
        float *v;

        v=(float *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(float)));
        if (!v) nrerror("allocation failure in vector()");
        return v-nl+NR_END;
}

int *ivector(long nl, long nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
        int *v;

        v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
        if (!v) nrerror("allocation failure in ivector()");
        return v-nl+NR_END;
}

unsigned char *cvector(long nl, long nh)
/* allocate an unsigned char vector with subscript range v[nl..nh] */
{
        unsigned char *v;

        v=(unsigned char *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(unsigned char)));
        if (!v) nrerror("allocation failure in cvector()");
        return v-nl+NR_END;
}

unsigned long *lvector(long nl, long nh)
/* allocate an unsigned long vector with subscript range v[nl..nh] */
{
        unsigned long *v;

        v=(unsigned long *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(long)));
        if (!v) nrerror("allocation failure in lvector()");
        return v-nl+NR_END;
}

double *dvector(long nl, long nh)
/* allocate a double vector with subscript range v[nl..nh] */
{
        double *v;

        v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
        if (!v) nrerror("allocation failure in dvector()");
        return v-nl+NR_END;
}

float **matrix(long nrl, long nrh, long ncl, long nch)
/* allocate a float matrix with subscript range m[nrl..nrh][ncl..nch] */
{
        long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        float **m;

        /* allocate pointers to rows */
        m=(float **) malloc((size_t)((nrow+NR_END)*sizeof(float*)));
        if (!m) nrerror("allocation failure 1 in matrix()");
        m += NR_END;
        m -= nrl;

        /* allocate rows and set pointers to them */
        m[nrl]=(float *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(float)));
        if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
        m[nrl] += NR_END;
        m[nrl] -= ncl;

        for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

        /* return pointer to array of pointers to rows */
        return m;
}

double **dmatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */
{
        long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        double **m;

        /* allocate pointers to rows */
        m=(double **) malloc((size_t)((nrow+NR_END)*sizeof(double*)));
        if (!m) nrerror("allocation failure 1 in matrix()");
        m += NR_END;
        m -= nrl;

        /* allocate rows and set pointers to them */
        m[nrl]=(double *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
        if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
        m[nrl] += NR_END;
        m[nrl] -= ncl;

        for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

        /* return pointer to array of pointers to rows */
        return m;
}

int **imatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a int matrix with subscript range m[nrl..nrh][ncl..nch] */
{
        long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        int **m;

        /* allocate pointers to rows */
        m=(int **) malloc((size_t)((nrow+NR_END)*sizeof(int*)));
        if (!m) nrerror("allocation failure 1 in matrix()");
        m += NR_END;
        m -= nrl;


        /* allocate rows and set pointers to them */
        m[nrl]=(int *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(int)));
        if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
        m[nrl] += NR_END;
        m[nrl] -= ncl;

        for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

        /* return pointer to array of pointers to rows */
        return m;
}

float **submatrix(float **a, long oldrl, long oldrh, long oldcl, long oldch,
        long newrl, long newcl)
/* point a submatrix [newrl..][newcl..] to a[oldrl..oldrh][oldcl..oldch] */
{
        long i,j,nrow=oldrh-oldrl+1,ncol=oldcl-newcl;
        float **m;

        /* allocate array of pointers to rows */
        m=(float **) malloc((size_t) ((nrow+NR_END)*sizeof(float*)));
        if (!m) nrerror("allocation failure in submatrix()");
        m += NR_END;
        m -= newrl;

        /* set pointers to rows */
        for(i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+ncol;

        /* return pointer to array of pointers to rows */
        return m;
}

float **convert_matrix(float *a, long nrl, long nrh, long ncl, long nch)
/* allocate a float matrix m[nrl..nrh][ncl..nch] that points to the matrix
declared in the standard C manner as a[nrow][ncol], where nrow=nrh-nrl+1
and ncol=nch-ncl+1. The routine should be called with the address
&a[0][0] as the first argument. */
{
        long i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1;
        float **m;

        /* allocate pointers to rows */
        m=(float **) malloc((size_t) ((nrow+NR_END)*sizeof(float*)));
        if (!m) nrerror("allocation failure in convert_matrix()");
        m += NR_END;
        m -= nrl;

        /* set pointers to rows */
        m[nrl]=a-ncl;
        for(i=1,j=nrl+1;i<nrow;i++,j++) m[j]=m[j-1]+ncol;
        /* return pointer to array of pointers to rows */
        return m;
}

float ***f3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
/* allocate a float 3tensor with range t[nrl..nrh][ncl..nch][ndl..ndh] */
{
        long i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1,ndep=ndh-ndl+1;
        float ***t;

        /* allocate pointers to pointers to rows */
        t=(float ***) malloc((size_t)((nrow+NR_END)*sizeof(float**)));
        if (!t) nrerror("allocation failure 1 in f3tensor()");
        t += NR_END;
        t -= nrl;

        /* allocate pointers to rows and set pointers to them */
        t[nrl]=(float **) malloc((size_t)((nrow*ncol+NR_END)*sizeof(float*)));
        if (!t[nrl]) nrerror("allocation failure 2 in f3tensor()");
        t[nrl] += NR_END;
        t[nrl] -= ncl;

        /* allocate rows and set pointers to them */
        t[nrl][ncl]=(float *) malloc((size_t)((nrow*ncol*ndep+NR_END)*sizeof(float)));
        if (!t[nrl][ncl]) nrerror("allocation failure 3 in f3tensor()");
        t[nrl][ncl] += NR_END;
        t[nrl][ncl] -= ndl;

        for(j=ncl+1;j<=nch;j++) t[nrl][j]=t[nrl][j-1]+ndep;
        for(i=nrl+1;i<=nrh;i++) {
                t[i]=t[i-1]+ncol;
                t[i][ncl]=t[i-1][ncl]+ncol*ndep;
                for(j=ncl+1;j<=nch;j++) t[i][j]=t[i][j-1]+ndep;
        }

        /* return pointer to array of pointers to rows */
        return t;
}

void free_vector(float *v, long nl, long nh)
/* free a float vector allocated with vector() */
{
        free((FREE_ARG) (v+nl-NR_END));
}

void free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */
{
        free((FREE_ARG) (v+nl-NR_END));
}

void free_cvector(unsigned char *v, long nl, long nh)
/* free an unsigned char vector allocated with cvector() */
{
        free((FREE_ARG) (v+nl-NR_END));
}

void free_lvector(unsigned long *v, long nl, long nh)
/* free an unsigned long vector allocated with lvector() */
{
        free((FREE_ARG) (v+nl-NR_END));
}

void free_dvector(double *v, long nl, long nh)
/* free a double vector allocated with dvector() */
{
        free((FREE_ARG) (v+nl-NR_END));
}

void free_matrix(float **m, long nrl, long nrh, long ncl, long nch)
/* free a float matrix allocated by matrix() */
{
        free((FREE_ARG) (m[nrl]+ncl-NR_END));
        free((FREE_ARG) (m+nrl-NR_END));
}

void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch)
/* free a double matrix allocated by dmatrix() */
{
        free((FREE_ARG) (m[nrl]+ncl-NR_END));
        free((FREE_ARG) (m+nrl-NR_END));
}

void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch)
/* free an int matrix allocated by imatrix() */
{
        free((FREE_ARG) (m[nrl]+ncl-NR_END));
        free((FREE_ARG) (m+nrl-NR_END));
}

void free_submatrix(float **b, long nrl, long nrh, long ncl, long nch)
/* free a submatrix allocated by submatrix() */
{
        free((FREE_ARG) (b+nrl-NR_END));
}

void free_convert_matrix(float **b, long nrl, long nrh, long ncl, long nch)
/* free a matrix allocated by convert_matrix() */
{
        free((FREE_ARG) (b+nrl-NR_END));
}

void free_f3tensor(float ***t, long nrl, long nrh, long ncl, long nch,
        long ndl, long ndh)
/* free a float f3tensor allocated by f3tensor() */
{
        free((FREE_ARG) (t[nrl][ncl]+ndl-NR_END));
        free((FREE_ARG) (t[nrl]+ncl-NR_END));
        free((FREE_ARG) (t+nrl-NR_END));
}



#include <math.h>

void simp1(float **a, int mm, int ll[], int nll, int iabf, int *kp,
        float *bmax)
{
        int k;
        float test;

        if (nll <= 0)
                *bmax=0.0;
        else {
                *kp=ll[1];
                *bmax=a[mm+1][*kp+1];
                for (k=2;k<=nll;k++) {
                        if (iabf == 0)
                                test=a[mm+1][ll[k]+1]-(*bmax);
                        else
                                test=fabs(a[mm+1][ll[k]+1])-fabs(*bmax);
                        if (test > 0.0) {
                                *bmax=a[mm+1][ll[k]+1];
                                *kp=ll[k];
                        }
                }
        }
}

#define EPS 1.0e-6

void simp2(float **a, int m, int n, int *ip, int kp)
{
        int k,i;
        float qp,q0,q,q1;

        *ip=0;
        for (i=1;i<=m;i++)
                if (a[i+1][kp+1] < -EPS) break;
        if (i>m) return;
        q1 = -a[i+1][1]/a[i+1][kp+1];
        *ip=i;
        for (i=*ip+1;i<=m;i++) {
                if (a[i+1][kp+1] < -EPS) {
                        q = -a[i+1][1]/a[i+1][kp+1];
                        if (q < q1) {
                                *ip=i;
                                q1=q;
                        } else if (q == q1) {
                                for (k=1;k<=n;k++) {
                                        qp = -a[*ip+1][k+1]/a[*ip+1][kp+1];
                                        q0 = -a[i+1][k+1]/a[i+1][kp+1];
                                        if (q0 != qp) break;
                                }
                                if (q0 < qp) *ip=i;
                        }
                }
        }
}
#undef EPS

void simp3(float **a, int i1, int k1, int ip, int kp)
{
        int kk,ii;
        float piv;

        piv=1.0/a[ip+1][kp+1];
        for (ii=1;ii<=i1+1;ii++)
                if (ii-1 != ip) {
                        a[ii][kp+1] *= piv;
                        for (kk=1;kk<=k1+1;kk++)
                                if (kk-1 != kp)
                                        a[ii][kk] -= a[ip+1][kk]*a[ii][kp+1];
                }
        for (kk=1;kk<=k1+1;kk++)
                if (kk-1 != kp) a[ip+1][kk] *= -piv;
        a[ip+1][kp+1]=piv;
}

#define NRANSI
#define EPS 1.0e-6
#define FREEALL free_ivector(l3,1,m);free_ivector(l1,1,n+1);

void simplx(float **a, int m, int n, int m1, int m2, int m3, int *icase,
        int izrov[], int iposv[])
{
        void simp1(float **a, int mm, int ll[], int nll, int iabf, int *kp,
                float *bmax);
        void simp2(float **a, int m, int n, int *ip, int kp);
        void simp3(float **a, int i1, int k1, int ip, int kp);
        int i,ip,is,k,kh,kp,nl1;
        int *l1,*l3;
        float q1,bmax;

        if (m != (m1+m2+m3)) nrerror("Bad input constraint counts in simplx");
        l1=ivector(1,n+1);
        l3=ivector(1,m);
        nl1=n;
        for (k=1;k<=n;k++) l1[k]=izrov[k]=k;
        for (i=1;i<=m;i++) {
                if (a[i+1][1] < 0.0) nrerror("Bad input tableau in simplx");
                iposv[i]=n+i;
        }
        if (m2+m3) {
                for (i=1;i<=m2;i++) l3[i]=1;
                for (k=1;k<=(n+1);k++) {
                        q1=0.0;
                        for (i=m1+1;i<=m;i++) q1 += a[i+1][k];
                        a[m+2][k] = -q1;
                }
                for (;;) {
                        simp1(a,m+1,l1,nl1,0,&kp,&bmax);
                        if (bmax <= EPS && a[m+2][1] < -EPS) {
                                *icase = -1;
                                FREEALL return;
                        } else if (bmax <= EPS && a[m+2][1] <= EPS) {
                                for (ip=m1+m2+1;ip<=m;ip++) {
                                        if (iposv[ip] == (ip+n)) {
                                                simp1(a,ip,l1,nl1,1,&kp,&bmax);
                                                if (bmax > EPS)
                                                        goto one;
                                        }
                                }
                                for (i=m1+1;i<=m1+m2;i++)
                                        if (l3[i-m1] == 1)
                                                for (k=1;k<=n+1;k++)
                                                        a[i+1][k] = -a[i+1][k];
                                break;
                        }
                        simp2(a,m,n,&ip,kp);
                        if (ip == 0) {
                                *icase = -1;
                                FREEALL return;
                        }
        one:	simp3(a,m+1,n,ip,kp);
                        if (iposv[ip] >= (n+m1+m2+1)) {
                                for (k=1;k<=nl1;k++)
                                        if (l1[k] == kp) break;
                                --nl1;
                                for (is=k;is<=nl1;is++) l1[is]=l1[is+1];
                        } else {
                                kh=iposv[ip]-m1-n;
                                if (kh >= 1 && l3[kh]) {
                                        l3[kh]=0;
                                        ++a[m+2][kp+1];
                                        for (i=1;i<=m+2;i++)
                                                a[i][kp+1] = -a[i][kp+1];
                                }
                        }
                        is=izrov[kp];
                        izrov[kp]=iposv[ip];
                        iposv[ip]=is;
                }
        }
        for (;;) {
                simp1(a,0,l1,nl1,0,&kp,&bmax);
                if (bmax <= EPS) {
                        *icase=0;
                        FREEALL return;
                }
                simp2(a,m,n,&ip,kp);
                if (ip == 0) {
                        *icase=1;
                        FREEALL return;
                }
                simp3(a,m,n,ip,kp);
                is=izrov[kp];
                izrov[kp]=iposv[ip];
                iposv[ip]=is;
        }
}
#undef EPS
#undef FREEALL
#undef NRANSI
