#include "nurbs.h"
#include <vector>
#include <iostream>


void initBinomialArray(Array2D<int> &b, int s)
{
    b.setSize(s,s);
    b.fill(0);
    for (int n = 0; n < s; n++) {
        for (int k = 0; k <= n; k++) {
            b(n,k) = binomial(n,k);
        }
    }

}

template<class T> class BernsteinCache
{
protected:
    Array2D<Polynom<T> > polynoms;

public:
    BernsteinCache(int n)
    {
        polynoms.setSize(n+1,n+1);
        for (int my_n = 0; my_n <= n; my_n++) {
            for (int i = 0; i <= my_n; i++) {
                polynoms(i, my_n) = bernsteinPolynom<T>(i, my_n);
            }
        }
    }
    inline Polynom<T> getPolynom(int i, int n)
    {
        if (n > polynoms.width()) {
            return bernsteinPolynom<T>(i, n);
        }
        return polynoms(i,n);

    }

};

template<> Polynom<double> bernsteinPolynomCache(int i, int n)
{
    static BernsteinCache<double> cache(10);
    return cache.getPolynom(i,n);
}

KnotVector::KnotVector(int p)
{
    m_compiled = false;
    m_knotvector.setSize(p);
    m_lastresult.m_index = -1;
    m_isbezier = false;
}

KnotVector::KnotVector(const Array<UTIFLOAT>& pos)
{
    m_compiled = false;
    m_isbezier = false;
    m_knotvector.setSize(pos.size());
    for (int i = 0; i < pos.size(); i++)
    {
        m_knotvector[i] = pos[i];
    }
}

KnotVector* KnotVector::createBezierKnotVector(int p)
{
   KnotVector*k = new KnotVector((p+1)*2.0);
   for (int i = 0; i <= p; i++) {
      k->setEntry(i, 0.0);
   }
   for (int i = p+1; i <= p*2+1; i++) {
      k->setEntry(i, 1.0);
   }
   k->setBezier(true);
   k->print();
   k->compile();
   return k;
}

void KnotVector::setBezier(bool b)
{ CHECKEXISTANCE;
    utiassert(m_compiled == false);
    m_isbezier = b;



}
bool KnotVector::getBezier() const
{ CHECKEXISTANCE;
    return m_isbezier;

}

void KnotVector::setEntry(int i, UTIFLOAT value)
{ CHECKEXISTANCE;
    utiassert(m_compiled == false);
    m_knotvector[i] = value;
}

int KnotVector::size() const
{ CHECKEXISTANCE;
    return m_knotvector.size();
}

int KnotVector::size_single() const
{ CHECKEXISTANCE;
    utiassert(m_compiled == true);
    return m_compile_knotvectorpos.size();
}

UTIFLOAT KnotVector::getSingle(int i) const
{ CHECKEXISTANCE;
    utiassert(m_compiled == true);
    return m_compile_knotvectorpos[i];
}

int KnotVector::getSingleMultiplicy(int i) const
{ CHECKEXISTANCE;
    utiassert(m_compiled == true);
    return m_compile_count[i];
}

bool KnotVector::isCompiled() const
{ CHECKEXISTANCE;
    return m_compiled;
}

void KnotVector::compile()
{ CHECKEXISTANCE;
    utiassert(m_compiled == false);
    std::vector<UTIFLOAT> count;
    std::vector<UTIFLOAT> knotvectorpos;
    std::vector<UTIFLOAT> knotvectorindex;

    int lastindex = 0;
    double lastvalue = m_knotvector[0];
    for (int i = 1; i < m_knotvector.size(); i++)
    {
        utiassert(m_knotvector[i] >= lastvalue);
        lastvalue = m_knotvector[i];

        if (m_knotvector[i] != m_knotvector[lastindex])
        {
            count.push_back(i-lastindex);
            knotvectorpos.push_back(m_knotvector[lastindex]);
            knotvectorindex.push_back(lastindex);
            lastindex = i;
        }
    }
    count.push_back(m_knotvector.size()-lastindex);
    knotvectorpos.push_back(m_knotvector[lastindex]);
    knotvectorindex.push_back(lastindex);
    m_compile_count.setSize(count.size());
    m_compile_index.setSize(count.size());
    m_compile_knotvectorpos.setSize(count.size());


    for (int i = 0; i< m_compile_count.size(); i++)
    {
        m_compile_count[i] = count[i];
        m_compile_index[i] = knotvectorindex[i];
        m_compile_knotvectorpos[i] = knotvectorpos[i];
    }


    m_compiled = true;


}


SKnotVectorResult KnotVector::findKnotVectorPositionSimple(UTIFLOAT v) const
{ CHECKEXISTANCE;
    SKnotVectorResult result;
    result.m_index = -1;
    for (int i = 0; i < m_compile_knotvectorpos.size(); i++)
    {
        UTIFLOAT value = m_compile_knotvectorpos[i];
        int pos = i;
        if (v <= value)
        {
            if (v != value)
            {
                pos--;

            }
            result.m_a = value;
            result.m_b = m_compile_knotvectorpos[pos+1];
            result.m_a_count = m_compile_count[pos];
            result.m_b_count = m_compile_count[pos+1];
            result.m_index = m_compile_index[pos+1]-1;
        }

    }
    utiassert(result.m_index != -1);
    return result;
}

UTIFLOAT KnotVector::getMinElement() const
{ CHECKEXISTANCE;
   utiassert(m_compiled == true);
   return m_compile_knotvectorpos[0];

}
UTIFLOAT KnotVector::getMaxElement() const
{ CHECKEXISTANCE;
    utiassert(m_compiled == true);
return m_compile_knotvectorpos[m_compile_knotvectorpos.size()-1];
}

void KnotVector::findKnotVectorPosition(UTIFLOAT v, SKnotVectorResult& result) const
{ CHECKEXISTANCE;
    utiassert(m_compiled == true);
    if (m_lastresult.m_index != -1 && m_lastresult.m_a <= v && v <= m_lastresult.m_b)
    {
        result = m_lastresult;
        return;
    }

    //binary search

    int start = 0;
    int end = m_compile_count.size();
    if (v >= m_compile_knotvectorpos[end-1]) {
        start = end-2;
        end = end-1;
    }
    while (true)
    {
        if (start == end-1)
        {
            break;
        }
        int mid = umax(start+1,(end-start) / 2);
        UTIFLOAT value  = m_compile_knotvectorpos[mid];
        if (v < value)
        {
            end = mid;
        }
        else
        {
            start = mid;
        }
    }
    utiassert(start == end-1);
    utiassert(end < m_compile_knotvectorpos.size());

    SKnotVectorResult& lastresult = const_cast<SKnotVectorResult&>(m_lastresult);

    lastresult.m_a = m_compile_knotvectorpos[start];
    lastresult.m_b = m_compile_knotvectorpos[end];
    lastresult.m_index = m_compile_index[end]-1;
    lastresult.m_a_count = m_compile_count[start];
    lastresult.m_b_count = m_compile_count[end];
    utiassert(m_compile_knotvectorpos[start] <= v && v <= m_compile_knotvectorpos[end]);
    result = m_lastresult;
}
/*
UTIFLOAT KnotVector::getBasisFunctionSimple(int i, int p, UTIFLOAT u)
{
    return getBasisFunctionSimple(i,p, p, u);
}
*/
UTIFLOAT KnotVector::getBasisFunctionSimple(int i, int p, UTIFLOAT u) const
{ CHECKEXISTANCE;
    utiassert(p >= 0);
    int m = m_knotvector.size()-1;
    if ((i == 0 && u == m_knotvector[0]) ||
            (i == m-p-1 && u == m_knotvector[m]))
    {
        return 1.0;
    }
    /*
        if  (u < m_knotvector[i] || u >= m_knotvector[i+p+1])
        {
            return 0.0;
        }*/
    if (p == 0)
    {
        UTIFLOAT ui = m_knotvector[i];
        UTIFLOAT ui1 = m_knotvector[i+1];

        /* if (i == m-p-1 && m_knotvector[m] == u) {
            return 1.0;
         }*/

        if (ui <= u && u < ui1)
        {
            return 1.0;
        }
        else
        {
            return 0.0;
        }
    }
    else
    {
        UTIFLOAT ui = m_knotvector[i];
        UTIFLOAT ui1 = m_knotvector[i+1];
        UTIFLOAT uip = m_knotvector[i+p];
        UTIFLOAT uip1 = m_knotvector[i+p+1];
        UTIFLOAT res = 0.0;

        if (uip != ui)
        {
            res += (u-ui)/(uip-ui)*getBasisFunctionSimple(i, p-1, u);
        }
        if (uip1 != ui1)
        {
            res += (uip1-u) / (uip1-ui1)*getBasisFunctionSimple(i+1, p-1, u);
        }

        return res;
    }
}

void KnotVector::dersBasisFunctionSimple(int i, int p, UTIFLOAT u, Array<UTIFLOAT>& ders, int derv) const
{ CHECKEXISTANCE;
    for (int n = 0; n < derv; n++)
    {
        ders[n] = dersBasisFunctionSimple(i,p,u,n);
    }
}

UTIFLOAT KnotVector::dersBasisFunctionSimple(int i, int p, UTIFLOAT u, int n) const
{ CHECKEXISTANCE;
    //int n = ders.width()-1;
    if (n == 0)
    {
        return getBasisFunctionSimple(i, p, u);
    }
    else
    {
        UTIFLOAT ui = m_knotvector[i];
        UTIFLOAT ui1 = m_knotvector[i+1];
        UTIFLOAT uip = m_knotvector[i+p];
        UTIFLOAT uip1 = m_knotvector[i+p+1];

        UTIFLOAT res = 0.0;

        if (uip != ui)
        {
            res += p/(uip-ui)*dersBasisFunctionSimple(i, p-1, u, n-1);
        }
        if (uip1 != ui1)
        {
            res += p / (uip1-ui1)*dersBasisFunctionSimple(i+1, p-1, u, n-1);
        }

        return res;
    }
}

UTIFLOAT KnotVector::getBasisFunction(int i, int p, UTIFLOAT u) const
{ CHECKEXISTANCE;
 static Array<UTIFLOAT> m_left;
    static Array<UTIFLOAT> m_right;
    static Array<UTIFLOAT> m_N;
    if (m_left.size() < p+1)
    {
        m_left.setSize(p+1);
        m_right.setSize(p+1);
        m_N.setSize(p+1);
    };
    UTIFLOAT saved, temp, Uleft, Uright;
    int m = m_knotvector.size()-1;
    if ((i == 0 && u == m_knotvector[0]) ||
            (i == m-p-1 && u == m_knotvector[m]))
    {
        return 1.0;
    }
    if  (u < m_knotvector[i] || u >= m_knotvector[i+p+1])
    {
        return 0.0;
    }
    for (int j =  0; j <= p; j++)
    {
        if (u >= m_knotvector[i+j] && u < m_knotvector[i+j+1])
        {
            m_N[j] = 1.0;
        }
        else
        {
            m_N[j] = 0.0;
        }
    }
    for (int k = 1; k <= p; k++)
    {
        if (m_N[0] == 0.0)
        {
            saved = 0.0;
        }
        else
        {
            saved = ((u-m_knotvector[i])*m_N[0])/(m_knotvector[i+k]-m_knotvector[i]);
        }
        for (int j = 0; j < p-k+1; j++)
        {
            Uleft = m_knotvector[i+j+1];
            Uright = m_knotvector[i+j+k+1];
            if  (m_N[j+1] ==  0.0)
            {
                m_N[j] = saved;
                saved = 0.0;
            }
            else
            {
                temp = m_N[j+1]/(Uright-Uleft);
                m_N[j] = saved + (Uright-u)*temp;
                saved = (u-Uleft)*temp;
            }
        }
    }

    //utiassert(absf(getBasisFunctionSimple(i, p, u)-m_N[0]) < 0.01);

    return m_N[0];
}

void KnotVector::dersBasisFunction(int i, int p, UTIFLOAT u, Array<UTIFLOAT>& ders, int n) const
{ CHECKEXISTANCE;
    //int n = derv-1;
    static Array<UTIFLOAT> m_ND;
    if (m_ND.size() < n)
    {
        m_ND.setSize(n);
    }
    static Array2D<UTIFLOAT> m_ndu;
    if (m_ndu.width() < (p+1) || m_ndu.height() < (p+1))
    {
        m_ndu.setSize(p+1, p+1);
    }
    UTIFLOAT saved, temp, Uleft, Uright;
    if  (u < m_knotvector[i] || u >= m_knotvector[i+p+1])
    {
        for (int k = 0; k <= n; k++)
        {
            ders[k] =0.0;
        }
        return;
    }
    for  (int j = 0; j <= p; j++)
    {
        if (u >= m_knotvector[i+j] && u < m_knotvector[i+j+1])
        {
            m_ndu(j,0) = 1.0;
        }
        else
        {
            m_ndu(j,0) = 0.0;
        }
    }
    for (int k = 1; k <= p; k++)
    {
        if (m_ndu(0,k-1) == 0.0)
        {
            saved = 0.0;
        }
        else
        {
            saved = ((u-m_knotvector[i])*m_ndu(0,k-1))/(m_knotvector[i+k]-m_knotvector[i]);
        }
        for (int j = 0; j < p-k+1; j++)
        {
            Uleft = m_knotvector[i+j+1];
            Uright = m_knotvector[i+j+k+1];
            if  (m_ndu(j+1,k-1) ==  0.0)
            {
                m_ndu(j,k) = saved;
                saved = 0.0;
            }
            else
            {
                temp = m_ndu(j+1, k-1)/(Uright-Uleft);
                m_ndu(j,k) = saved + (Uright-u)*temp;
                saved = (u-Uleft)*temp;
            }
        }
    }
    ders[0] = m_ndu(0,p);
    for (int k = 1; k <= n; k++)
    {
        for (int j = 0; j <= k; j++)
        {
            m_ND[j] = m_ndu(j, p-k);
        }
        for (int jj = 1; jj <= k; jj++)
        {
            if (m_ND[0] == 0.0)
            {
                saved = 0.0;
            }
            else
            {
                saved = m_ND[0] / (m_knotvector[i+p-k+jj]-m_knotvector[i]);
            }
            for (int j = 0; j < k -jj +1; j++)
            {
                Uleft=  m_knotvector[i+j+1];
                Uright = m_knotvector[i+j+p+jj+1];
                if (m_ND[j+1] == 0.0)
                {
                    m_ND[j] = (p-k+jj)*saved;
                    saved = 0.0;
                }
                else
                {
                    temp = m_ND[j+1]/(Uright-Uleft);
                    m_ND[j] = (p-k+jj)*(saved-temp);
                    saved = temp;
                }
            }
        }
        ders[k] = m_ND[0];
    }
}

void KnotVector::calcBasisFunctions(int i, UTIFLOAT u, Array<UTIFLOAT>& N, int p) const
{
    CHECKEXISTANCE;
    static Array<UTIFLOAT> m_left;
    static Array<UTIFLOAT> m_right;
    static Array<UTIFLOAT> m_N;
  //  int p = N.size()-1;
    utiassert(N.size() >= p+1);
    if (m_left.size() < p+1)
    {
        m_left.setSize(p+1);
        m_right.setSize(p+1);
        m_N.setSize(p+1);
    };

    UTIFLOAT temp;
    UTIFLOAT saved;
    N[0] = 1.0f;
    for (int j = 1; j <= p; j++)
    {
        m_left[j] = u-m_knotvector[i+1-j];
        m_right[j] = m_knotvector[i+j]-u;
        saved = 0.0;
        for (int r = 0; r < j; r++)
        {
            temp = N[r]/(m_right[r+1]+m_left[j-r]);
            N[r] = saved+m_right[r+1]*temp;
            saved = m_left[j-r]*temp;
        }
        N[j] = saved;
    }
#ifndef NDEBUG
    for (int j = 0; j <= p; j++)
    {
        UTIFLOAT f = getBasisFunction(j+i-p, p, u);
        utiassert(absf(f-N[j]) < 0.01);
    }
#endif
}

void KnotVector::dersBasisFunctions(int i, UTIFLOAT u, Array2D<UTIFLOAT> &ders, int n, int p) const
{ CHECKEXISTANCE;
    //int n = ders.width()-1;
   // int p = ders.height()-1;
   static Array<UTIFLOAT> m_left;
    static Array<UTIFLOAT> m_right;
    static Array<UTIFLOAT> m_N;
    static Array2D<UTIFLOAT> m_ndu;
    static Array2D<UTIFLOAT> m_a;
    if (m_left.size() < p+1)
    {
        m_left.setSize(p+1);
        m_right.setSize(p+1);
        m_N.setSize(p+1);
    };

    utiassert(ders.width() >= (n+1));
    utiassert(ders.height() >= (p+1));
    UTIFLOAT temp;
    UTIFLOAT saved;
    UTIFLOAT d;
    int rk, pk, s1,s2,j1, j2;
    if (m_ndu.width() < (p+1) || m_ndu.height() < (p+1))
    {
        m_ndu.setSize(p+1, p+1);
    }
    if (m_a.width() < 2 || m_a.height() < (p+1))
    {
        m_a.setSize(2, (p+1));
    }
    m_ndu(0,0) = 1.0;

    for (int j = 1; j <= p; j++)
    {
        m_left[j] = u-m_knotvector[i+1-j];
        m_right[j] = m_knotvector[i+j]-u;
        saved = 0.0;
        for (int r = 0; r < j; r++)
        {
            m_ndu(j,r) = (m_right[r+1]+m_left[j-r]);
            temp = m_ndu(r,j-1)/m_ndu(j,r);

            m_ndu(r,j) = saved+m_right[r+1]*temp;

            saved = m_left[j-r]*temp;
        }
        m_ndu(j,j) = saved;
    }
    for (int j = 0; j <=p; j++)
    {
        ders(0,j) = m_ndu(j,p);
    }
    for (int r = 0; r <= p; r++)
    {
        s1=0;
        s2=1;
        m_a(0,0) = 1.0;
        for (int k = 1; k <= n; k++)
        {
            d = 0.0;

            rk = r-k;
            pk = p-k;
            if (r >= k)
            {
                m_a(s2,0) = m_a(s1,0)/m_ndu(pk+1,rk);
                d = m_a(s2,0)*m_ndu(rk, pk);
            }
            if (rk >= -1)
            {
                j1 = 1;
            }
            else
            {
                j1 = -rk;
            }

            if (r-1 <= pk)
            {
                j2 = k-1;
            }
            else
            {
                j2 = p-r;
            }
            for (int j=j1; j <= j2; j++)
            {
                m_a(s2, j) = (m_a(s1,j)-m_a(s1,j-1))/m_ndu(pk+1,rk+j);
                d += m_a(s2,j)*m_ndu(rk+j, pk);
            }
            if (r <= pk)
            {
                m_a(s2,k) = (-m_a(s1, k-1)/m_ndu(pk+1, r));
                d += m_a(s2,k)*m_ndu(r,pk);
            }
            ders(k,r)=d;
            int tj=s1;
            s1=s2;
            s2=tj;

        }
    }
    int r=p;
    for (int k = 1; k <= n; k++)
    {
        for (int j = 0; j <= p; j++)
        {
            ders(k,j) *= r;

        }
        r +=  (p-k);
    }
}





void KnotVector::print() const
{ CHECKEXISTANCE;
    std::wcout << "KnotVector (";
    for (int i = 0; i <m_knotvector.size(); i++)
    {
        CUtiString str;
        str.FloatToStr(m_knotvector[i], 0,2);
        std::wcout << str;
        if (i != m_knotvector.size()-1)
        {
            std::wcout << ", ";
        }
    }
    std::wcout << ")" << std::endl;
}


void KnotVector::copyToKnot(int ownstart, int targetstart, int count, KnotVector* target) const
{ CHECKEXISTANCE;
    utiassert(target->m_compiled == false);
    for (int i = 0; i < count; i++)
    {
        target->m_knotvector[i+targetstart] = m_knotvector[i+ownstart];
    }
}






