/***************************************************************************
 *   Copyright (C) 2017 by Staud                                           *
 *   michael.staud@uni-ulm.de                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef UTILISP_H
#define UTILISP_H

#include "Uti3DCore.h"
#include "utiadvarray.h"
#include "utiarray.h"
#include "lispnames.h"

#include "utimemmanager.h"
#include <limits>

namespace Lisp {

enum TListCellType {tctNone, tctFloat, tctInteger, tctReturn, tctAtom, tctCons, tctString, tctArray,tctMultiArray, tctDynamicArray, tctTrue,  tctLambda, tctProc, tctNil};


class CLispFunction;




struct LObject;


class LCell : public ReferenceUtiOB
{
public:

    TListCellType m_type;
    inline LCell() {m_type = tctNone;}
    virtual bool l_equal(LCell* other) const;
    virtual bool l_eql(LCell* other) const;
    virtual int l_hash() const;
    virtual bool l_smaller(LCell* other) const;
    virtual LCell* l_createCopy() const;
    virtual LObject l_eval() const;
    virtual CUtiString l_printAligned() const {return l_print();}
    virtual CUtiString l_print() const {return UTISTR("<unknown>");};
    virtual int l_typeof() const {return LISP_symbol;};
    virtual bool l_typep(int type) const {return false;};


    UTI_MEMORY_POOL_DECL(LCell, 65536, SMemoryPool_BaseBehaviour);
};

LCell *ReadCell(CStream & Stream, int Flags, int version);
HResult WriteCell(CStream & Stream, LCell* cell, int Flags, int version);

LCell* xreadFact(CXNode* node, CWriteReadInterface* wri);
void xwriteFact(LCell* cell, CXNode* node, CWriteReadInterface* wri);

UTI_MEMORY_POOL_DECL_EXTERN(LCell, 65536, SMemoryPool_BaseBehaviour);

class LInteger;
class LArray;
class LBaseArray;
class LList;
class LFloat;
class LAtom;
class LString;
class LTrue;
class LFalse;
class LNil;
class LCons;

#ifndef NDEBUG
//#define DEBUGLISP
#endif

struct LObject
{
   uti_refptr<LCell> m_data;
#ifdef DEBUGLISP
   std::string m_content;
#endif
   inline TListCellType getType() const {return m_data->m_type;}
   inline LObject();
   inline LObject(int value);
   inline LObject(double value);
   inline LObject(float value);
   inline LObject(bool b);
   inline LObject(const CUtiString& str);
   inline LObject(const CUtiIDString& str);
   inline LObject(const LObject& str);
   inline LObject(LCell* cell) {m_data = cell;}
   inline LObject(const LCell* cell) {m_data = const_cast<LCell*>(cell);}
   inline LObject(LString* cell);
   inline LObject(const LString* cell);
   inline LObject(LInteger* cell);
   inline LObject(const LInteger* cell);
   inline LObject(LFloat* cell);
   inline LObject(const LFloat* cell);
   inline LObject(LList* cell);
   inline LObject(const LList* cell);
   inline LObject(LCons* cell);
   inline LObject(const LCons* cell);
   inline LObject(LBaseArray* cell);
   inline LObject(const LBaseArray* cell);
   inline LObject(LAtom* cell);
   inline LObject(const LAtom* cell);
   inline LObject(LTrue* cell);
   inline LObject(const LTrue* cell);
   inline LObject(LNil* cell);
   inline LObject(const LNil* cell);

   inline LObject& operator=(int value);
   inline LObject& operator=(float value);
   inline LObject& operator=(double value);
   inline LObject& operator=(bool b);
   inline LObject& operator=(const CUtiString& str);
   inline LObject& operator=(const CUtiIDString& str);
   inline LObject& operator=(const LObject& str);
   inline LObject& operator=(LCell* cell) {m_data = cell;return *this;}
   inline bool operator==(const LObject& obj) const
   {
       return m_data->l_equal(obj.m_data.get());
   }
   inline bool operator!=(const LObject& obj) const
   {
       return !m_data->l_equal(obj.m_data.get());
   }

   inline bool operator<(const LObject& obj) const
   {
       return m_data->l_smaller(obj.m_data.get());
   }
   inline bool operator>(const LObject& obj) const
   {
       return  obj.m_data->l_smaller(m_data.get());
   }
   inline bool operator>=(const LObject& obj) const
   {

       return operator==(obj) || operator>(obj);
   }
   inline bool operator<=(const LObject& obj) const
   {

       return operator==(obj) || operator<(obj);
   }
   inline LObject eval() const {return m_data->l_eval();}
   inline bool isArray() const {return m_data->m_type == tctArray|| m_data->m_type == tctMultiArray;}
   inline bool isConList() const;
   inline bool isNil() const;
   inline void setNil();
   inline bool isAtom() const;
   inline bool isString() const;
   inline bool isThisAtom(const CUtiIDString& str) const;
   inline CUtiIDString getAtom() const;
   inline bool isLambda()const
   {
       return m_data->m_type == tctLambda;
   }
   inline bool isNumber() const
   {
       return m_data->m_type == tctInteger || m_data->m_type == tctFloat;
   }
   inline double toFloat() const;
   inline int toInt() const;
   inline bool toBool() const;

   inline bool isBool() const;
   inline CUtiString toString() const;

   std::string createDebugData() const;
   inline CUtiString print() const {
       return m_data->l_print();
   }
   inline CUtiString printAligned() const {
       return m_data->l_printAligned();
   }
   inline int hashCode()
   {
       return m_data->hashCode();
   }
   struct iterator: public std::iterator<
           std::forward_iterator_tag,   // iterator_category
           long,                      // value_type
           long,                      // difference_type
           const long*,               // pointer
           long                       // reference
                         >
   {
       uti_refptr<LCell> m_data;
       iterator& operator=(const iterator& it);
       bool operator==(const iterator& it);
       bool operator!=(const iterator& it);
       iterator& operator++(int); //prefix increment
       iterator& operator++(); //prefix increment
       LObject operator*() const;


   };

   iterator begin();
   iterator end();
   iterator begin() const;
   iterator end() const;
   void toList(DynArray1D<LObject>& subids);

   inline void updateDebug()
   {

#ifdef DEBUGLISP
       static bool inDebugData = false;
       if (inDebugData) return;
       inDebugData = true;
       m_content = createDebugData();
       inDebugData = false;
#endif
   }
};


class LEnvironment : public ReferenceUtiOB
{
public:
    // inline LObject get(const CUtiIDString& str);
    // inline void set(const CUtiIDString& str, const LObject& obj);
    // inline void createVar(const CUtiIDString& str, const LObject& obj);
    std::map<CUtiIDString, int> m_symbolsmap;
    DynArray1D<LObject> m_symbols;
    uti_refptr<LEnvironment> m_parent;
};


class CLispFunction : public ReferenceUtiOB
{
public:
    virtual LObject call(const LObject& arguments)
    {
        return LObject();
    }
};


class LNumber : public LCell
{
public:
    inline LNumber() {};
};

class LispException : public Exception
{
public:
    LObject m_object;
    CUtiString m_data;
    LispException(LObject obj, const CUtiString& data)
    {
        m_data = data;
        m_object = obj;
        std::wcout << m_data << " in " << obj.print() << std::endl;

    }
};


class LGlobalEnvironment : public CUtiOB
{
public:
   uti_refptr<LEnvironment> m_global;
   uti_refptr<LEnvironment> m_toponstack;
   static LGlobalEnvironment* s_globalenvironment;
   static LGlobalEnvironment* init();
   LGlobalEnvironment()
   {
       m_global = new LEnvironment();
       m_toponstack = m_global.get();
   }

   inline LEnvironment* getTopStack() {return m_toponstack.get();}
   inline void setTopStack(LEnvironment* env) { m_toponstack = env;}
   inline void push() {
       LEnvironment* env = new LEnvironment();
       env->m_parent = m_toponstack;
       m_toponstack = env;
   }
   inline void pop() {
       m_toponstack = m_toponstack->m_parent;
   }
   inline void getVarData(const CUtiIDString& str, LEnvironment** env, int &index)
   {
       index = -1;
       LEnvironment* e = m_toponstack.get();
       while (e) {

            std::map<CUtiIDString, int>::iterator it = e->m_symbolsmap.find(str);
           if (it != e->m_symbolsmap.end()) {
               index = it->second;
               *env = e;
               return;
           }
           e = e->m_parent.get();
       }

   }

   inline void setVariable(const CUtiIDString& str, const LObject& data)
   {
       LEnvironment* env;
       int index;
       getVarData(str, &env, index);
       if (index == -1) throw LispException(LObject(), UTISTR("The variable with the name \"")+str.str()+UTISTR("\" was not found."));

       env->m_symbols[index] = data;
   }
   inline LObject getVariable(const CUtiIDString& str)
   {
       LEnvironment* env;
       int index;
       getVarData(str, &env, index);
       if (index == -1) throw LispException(LObject(), UTISTR("The variable with the name \"")+str.str()+UTISTR("\" was not found."));

       return env->m_symbols[index];

   }
   inline void createVariable(const CUtiIDString& str, const LObject& data)
   {
       int index = m_toponstack->m_symbols.push_back(data);
       m_toponstack->m_symbolsmap.insert(std::make_pair(str, index));
   }
};

class LInteger : public LNumber
{
public:
    int m_value;
    inline LInteger() { m_type = tctInteger; m_value = 0; }
    inline LInteger(int v) { m_type = tctInteger; m_value = v;}


    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return check_cast<LInteger*>(other)->m_value == m_value;}
    virtual int l_hash() const {return m_value;}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; if (other->m_type != m_type) throw LispException(other,UTISTR("Integer expected")); return m_value < check_cast<LInteger*>(other)->m_value;}
    virtual LCell* l_createCopy() const {
        LInteger* res = new LInteger();
        res->m_value = m_value;
        return res;
    }
    virtual CUtiString l_print() const {return CUtiString::intToStr(m_value);};


    virtual int l_typeof() const {return LISP_integer;};
    virtual bool l_typep(int type) const {return type == LISP_number || type == LISP_integer;};


    UTI_MEMORY_POOL_DECL(LInteger, 65536, SMemoryPool_BaseBehaviour);
};


UTI_MEMORY_POOL_DECL_EXTERN(LInteger, 65536, SMemoryPool_BaseBehaviour);



class LFloat : public LNumber
{
public:
    double m_value;
    inline LFloat(){  m_type = tctFloat;    m_value = 0; }
    inline LFloat(double v) { m_type = tctFloat;     m_value = v;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return check_cast<LFloat*>(other)->m_value == m_value;}
    virtual int l_hash() const {return *((int*)&m_value);}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; if (other->m_type != m_type) throw LispException(other,UTISTR("Float expected")); return m_value < check_cast<LFloat*>(other)->m_value;}
    virtual LCell* l_createCopy() const {
        LFloat* res = new LFloat();
        res->m_value = m_value;
        return res;
    }
    virtual CUtiString l_print() const {return CUtiString::floatToStr(m_value, 0,5);};
    virtual int l_typeof() const {return LISP_float;};
    virtual bool l_typep(int type) const {return type == LISP_number || type == LISP_float;};

    UTI_MEMORY_POOL_DECL(LFloat, 65536, SMemoryPool_BaseBehaviour);
};


UTI_MEMORY_POOL_DECL_EXTERN(LFloat, 65536, SMemoryPool_BaseBehaviour);

class LTrue : public LCell
{
public:
    LTrue(){m_type = tctTrue;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return true;}
    virtual int l_hash() const {return 0;}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; return false;}
    virtual LCell* l_createCopy() const {
        LTrue* res = new LTrue();
        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("t");}
    virtual int l_typeof() const {return LISP_t;};
    virtual bool l_typep(int type) const {return type == LISP_t;};

    UTI_MEMORY_POOL_DECL(LTrue, 65536, SMemoryPool_BaseBehaviour);
};



UTI_MEMORY_POOL_DECL_EXTERN(LTrue, 65536, SMemoryPool_BaseBehaviour);

class LReturn : public LCell
{
public:
    LObject m_value;
    CUtiIDString m_name;
    LReturn(){m_type = tctReturn;}

    virtual bool l_equal(LCell* other) const {assert(false);return false;}
    virtual int l_hash() const {return 0;}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; assert(false);return false;}
    virtual LCell* l_createCopy() const {
        assert(false);
        LReturn* res = new LReturn();

        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("return");}
    virtual int l_typeof() const {assert(false);return LISP_t;};
    virtual bool l_typep(int type) const {assert(false);return type == LISP_t;};
    UTI_MEMORY_POOL_DECL(LReturn, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LReturn, 65536, SMemoryPool_BaseBehaviour);
/*
class LFalse : public LCell
{
public:
    LFalse()  {m_type = tctFalse;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return true;}
    virtual int l_hash() const {return 0;}
    virtual bool l_smaller(LCell* other) const  {return false;}
    virtual LCell* l_createCopy() const {
        LFalse* res = new LFalse();
        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("f");}
    virtual int l_typeof() const {return LISP_t;};
    virtual bool l_typep(int type) const {return type == LISP_t;};

};*/


class LNil : public LCell
{
public:
    LNil()  {m_type = tctNil;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return true;}
    virtual int l_hash() const {return 0;}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; return false;}
    virtual LCell* l_createCopy() const {
        LNil* res = new LNil();
        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("nil");}
    virtual int l_typeof() const {return LISP_nil;}
    virtual bool l_typep(int type) const {return type == LISP_nil;}
    UTI_MEMORY_POOL_DECL(LNil, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LNil, 65536, SMemoryPool_BaseBehaviour);

class LAtom : public LCell
{
public:
    CUtiIDString m_value;
    inline LAtom(const CUtiIDString& d){ m_type = tctAtom;  m_value = d; }
    inline LAtom(){    m_type = tctAtom;  m_value.setID( CGameEngineNameStore::s_emptystring); }

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return check_cast<LAtom*>(other)->m_value == m_value;}
    virtual int l_hash() const {return m_value.getID();}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; return m_value.getID() < check_cast<LAtom*>(other)->m_value.getID();}
    virtual LCell* l_createCopy() const {
        LAtom* res = new LAtom();
        res->m_value = m_value;
        return res;
    }
    virtual CUtiString l_print() const {return m_value.str();}
    virtual LObject l_eval() const;
    virtual int l_typeof() const {return LISP_symbol;}
    virtual bool l_typep(int type) const {return type == LISP_symbol;}
    UTI_MEMORY_POOL_DECL(LAtom, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LAtom, 65536, SMemoryPool_BaseBehaviour);

inline LObject atom(const CUtiIDString & value)
{
    CUtiIDString str = value;
    LObject res;
    LAtom* ar = new LAtom(str);
    res.m_data = ar;
            return res;
}

class LProc : public LCell
{
public:
    uti_refptr<CLispFunction> m_value;
    LProc(CLispFunction* l) { m_type = tctProc; m_value = l;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return check_cast<LProc*>(other)->m_value == m_value;}
    virtual int l_hash() const {return m_value->hashCode();}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; return m_value.get() < check_cast<LProc*>(other)->m_value.get();}
    virtual LCell* l_createCopy() const {
        LProc* res = new LProc(m_value.get());
        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("<proc>");}
    virtual int l_typeof() const {return LISP_function;}
    virtual bool l_typep(int type) const {return type == LISP_function;}
UTI_MEMORY_POOL_DECL(LProc, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LProc, 65536, SMemoryPool_BaseBehaviour);

class LLambda : public LCell
{
public:
    LObject m_args;
    LObject m_sexp;
    uti_refptr<LEnvironment> m_environment;
    LLambda(const LObject& a, const LObject& sexp){m_type =tctLambda; m_args = a; m_sexp = sexp;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return (check_cast<LLambda*>(other)->m_args == m_args) &&
                (check_cast<LLambda*>(other)->m_sexp == m_sexp);}
    virtual int l_hash() const {return m_args.m_data->l_hash() + m_sexp.m_data->l_hash();}
    virtual bool l_smaller(LCell* other) const
    {
        if (m_type < other->m_type){return true; }if (m_type > other->m_type){return false;}
        if (other->m_type != m_type) throw LispException(other,UTISTR("Lambda expression expected"));
        LLambda* lo = check_cast<LLambda*>(other);
        if (m_args.m_data->l_smaller(lo->m_args.m_data.get())) return true;
        if (!m_args.m_data->l_equal(lo->m_args.m_data.get())) return false;
        if (m_sexp.m_data->l_smaller(lo->m_sexp.m_data.get())) return true;
        return false;
    }
    virtual LCell* l_createCopy() const
    {
        LLambda* res = new LLambda(m_args, m_sexp);
        res->m_environment = m_environment;

        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("(lambda ") + m_args.print() + UTISTR(" ") + m_sexp.print() +UTISTR(" )");}
    virtual int l_typeof() const {return LISP_function;}
    virtual bool l_typep(int type) const {return type == LISP_function;}

UTI_MEMORY_POOL_DECL(LLambda, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LLambda, 65536, SMemoryPool_BaseBehaviour);
class LString : public LCell
{
public:
    CUtiString m_value;
    LString() { m_type = tctString;}
    LString(const CUtiString& v) {m_type = tctString;m_value = v;}

    virtual bool l_equal(LCell* other) const {if (other->m_type != m_type) return false; return check_cast<LString*>(other)->m_value == m_value;}
    virtual int l_hash() const {return m_value.hashCode();}
    virtual bool l_smaller(LCell* other) const  {if (m_type < other->m_type)return true; if (m_type > other->m_type)return false; if (other->m_type != m_type) throw LispException(other,UTISTR("String expression expected"));
        return m_value < check_cast<LString*>(other)->m_value;}
    virtual LCell* l_createCopy() const {
        LString* res = new LString();
        res->m_value = m_value;
        return res;
    }
    virtual CUtiString l_print() const {return UTISTR("\"") + m_value + UTISTR("\"");}
    virtual int l_typeof() const {return LISP_l_string;}
    virtual bool l_typep(int type) const {return type == LISP_l_string;}


UTI_MEMORY_POOL_DECL(LString, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LString, 65536, SMemoryPool_BaseBehaviour);



class LBaseArray : public LCell
{
public:
    LBaseArray() {}
    virtual LObject getElement(int index)const=0;
    virtual void setElement(int index, const LObject& obj1)=0;
    virtual Array<int>* dimensionSize() {return NULL;}
    virtual int size()const=0;

    virtual bool l_equal(LCell* other) const
    {
#ifndef NDEBUG
    if (other->m_type != m_type && LObject(this).isArray() == LObject(other).isArray()) {
        assert(false);

    }
#endif
        if (other->m_type != m_type) return false;
        LBaseArray* al = check_cast<LBaseArray*>(other);
        int c = size();
        if (al->size() != c)return false;

        for (int i = 0; i < c; i++) {
            if (getElement(i) != al->getElement(i)) return false;
        }
        return true;

    }
    virtual int l_hash() const
    {
       int c = size();
       int hashvalue = 5381;
       for (int i = 0; i < c; i++) {
            hashvalue = ((hashvalue << 5) + hashvalue) + getElement(i).hashCode();
       };
       return hashvalue;
    }
    virtual bool l_smaller(LCell* other) const
    {
        if (m_type < other->m_type){return true;} if (m_type > other->m_type){return false;}
#ifndef NDEBUG
    if (other->m_type != m_type && LObject(this).isArray() == LObject(other).isArray()) {
        assert(false);

    }
#endif
        if (other->m_type != m_type) throw LispException(other,UTISTR("Array expression expected"));
        LBaseArray* al = check_cast<LBaseArray*>(other);
        int c = size();
        if (al->size() < c)return false;

        for (int i = 0; i < c; i++) {
            LObject obj1 = getElement(i);
            LObject obj2 = al->getElement(i);
            if (obj1 < obj2) return true;
            if (obj1 != obj2) return false;
        }
        return false;
    }
    virtual LCell* l_createCopy() const
    {
        assert(false);
        return NULL;
    }
};

class LFilterCallback
{
public:
    virtual bool eval(const LObject& obj) const=0;
};

class LMapCallback
{
public:
    virtual LObject eval(const LObject& obj) const=0;
};

class LList : public LCell
{
public:
    inline LList()
    {

    }
    virtual int l_length();
    virtual int l_count(const LObject& obj);
  //  int l_count_if(const LObject& obj, LObject& condition);
  //  int l_count_if_not(const LObject& obj, LObject& condition);
    virtual LObject l_car()=0;
   virtual  LObject l_cdr()=0;
    virtual void l_setcar(const LObject& obj)=0;
   virtual  void l_setcdr(const LObject& obj)=0;
    virtual LObject l_nth(int i);
    //LObject l_adjoin();
   // LObject l_sort();
    //LObject l_substitute(const LObject& newone, const LObject& old);
   // LObject l_substitute_if(const LObject& newone, const LObject& old, LObject& condition);
    //LObject l_substitute_if_not(const LObject& newone, const LObject& old, LObject& condition);
   // LObject l_reverse();
  //  LObject l_union(int i);
   // LObject l_intersection(int i);
   // LObject l_set_difference(int i);
    //LObject l_last();
   // LObject l_butlast();
    //LObject l_butlast(int i);
   // LObject l_nthcdr(int i);
   // LObject l_subseq(int start);
    virtual LObject l_subseq(int start, int end);
    virtual LObject l_filter(const LObject& obj, LFilterCallback *callback);
    virtual LObject l_map(const LObject& obj, LMapCallback *callback);


    virtual bool l_equal(LCell* other) const;
    virtual int l_hash() const;
    virtual bool l_smaller(LCell* other) const;
    virtual CUtiString l_print() const;
    virtual CUtiString l_printAligned() const;

    virtual LObject l_eval() const;
    virtual int l_typeof() const {return LISP_cons;}
    virtual bool l_typep(int type) const {return type == LISP_cons;}

UTI_MEMORY_POOL_DECL(LList, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LList, 65536, SMemoryPool_BaseBehaviour);

LObject push(const LObject& obj, LObject& list);
LObject pop(LObject& list);



int length(const LObject& obj);
inline int count(LObject obj, LObject o)
{
    if (obj.isNil())return 0;
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    return check_cast<LList*>(obj.m_data.get())->l_count(o);
}
inline LObject car(LObject obj)
{
    if (obj.isNil())return obj;
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    return check_cast<LList*>(obj.m_data.get())->l_car();
}
inline LObject cdr(LObject obj)
{
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    return check_cast<LList*>(obj.m_data.get())->l_cdr();
}
inline LObject nth(const LObject &obj, const LObject& pos)
{
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    return check_cast<LList*>(obj.m_data.get())->l_nth(pos.toInt());
    //return nth(obj, pos.toInt());

}

inline LObject nth(const LObject &obj, int pos)
{
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    return check_cast<LList*>(obj.m_data.get())->l_nth(pos);
}
inline LObject subseq(LObject obj, const LObject& pos)
{
   return subseq(obj, pos.toInt());
}

inline LObject subseq(LObject obj, int pos)
{
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    LList* list = check_cast<LList*>(obj.m_data.get());
    int l = list->l_length();
    return check_cast<LList*>(obj.m_data.get())->l_subseq(pos, l);
}
inline LObject subseq(LObject obj, int start, int end)
{
    if (!obj.isConList()) throw LispException(obj,UTISTR("Is not a sequence"));
    LList* list = check_cast<LList*>(obj.m_data.get());
    return list->l_subseq(start, end);
}
inline LObject first(LObject obj)
{
    return nth(obj, 0);
}

inline LObject second(LObject obj)
{
    return nth(obj, 1);
}


inline LObject third(LObject obj)
{
    return nth(obj, 2);
}

inline LObject last(LObject obj)
{
    int l = length(obj);
    return nth(obj,l-1);
}



class LCons : public LList
{
public:
    LObject m_head;
    LObject m_tail;
    inline LCons()
    {
        m_type = tctCons;

    }
    inline LCons(const LObject& obj)
    {
        m_type = tctCons;
        m_head = obj;
    }
    inline LCons(const LObject& obj, const LObject& t)
    {
        m_type = tctCons;
        m_head = obj;
        m_tail = t;
        assert(t.isConList());
    }
    LObject l_car(){return m_head; }
    LObject l_cdr(){return m_tail;}
    void l_setcar(const LObject& obj){m_head = obj;}
    void l_setcdr(const LObject& obj){m_tail = obj;}


    virtual LCell* l_createCopy() const;

UTI_MEMORY_POOL_DECL(LCons, 65536, SMemoryPool_BaseBehaviour);
};

UTI_MEMORY_POOL_DECL_EXTERN(LCons, 65536, SMemoryPool_BaseBehaviour);

inline LObject caar(const LObject& x) {return car(car(x));}
inline LObject cadr(const LObject& x) {return car(cdr(x));}
inline LObject cdar(const LObject& x) {return cdr(car(x));}
inline LObject cddr(const LObject& x) {return cdr(cdr(x));}
inline LObject caaar(const LObject& x) {return car(car(car(x)));}
inline LObject caadr(const LObject& x) {return car(car(cdr(x)));}
inline LObject cadar(const LObject& x) {return car(cdr(car(x)));}
inline LObject caddr(const LObject& x) {return car(cdr(cdr(x)));}
inline LObject cddar(const LObject& x) {return cdr(cdr(car(x)));}
inline LObject cdaar(const LObject& x) {return cdr(car(car(x)));}
inline LObject cdadr(const LObject& x) {return cdr(car(cdr(x)));}
inline LObject cdddr(const LObject& x) {return cdr(cdr(cdr(x)));}
inline LObject cadddr(const LObject& x) {return car(cdr(cdr(cdr(x))));}
inline LObject caddddr(const LObject& x) {return car(cdr(cdr(cdr(cdr(x)))));}
inline LObject cadddddr(const LObject& x) {return car(cdr(cdr(cdr(cdr(cdr(x))))));}

inline LObject t() {return LObject(true);}
inline LObject f() {return LObject(false);}
inline LObject nil() {return LObject();}

inline LObject cons(const LObject& head) {LObject obj; obj.m_data = new LCons(head); obj.updateDebug(); return obj;}
inline LObject cons(const LObject& head, const LObject& t1) {LObject obj; obj.m_data = new LCons(head, t1); obj.updateDebug();return obj;}

LObject list(const DynArray1D<LObject> & elements);
LObject list(const std::set<LObject> & elements);
void listToArray(const LObject& list, DynArray1D<LObject> & elements);
void listToArray(const LObject& list, std::set<LObject> & elements);
LObject lunion(const LObject& l1, const LObject& l2);
LObject lsort(const LObject& l1);
LObject reverse(const LObject& l1);
LObject lintersection(const LObject& l1, const LObject& l2);
LObject lset_difference(const LObject& l1, const LObject& l2);

inline LObject list(const LObject& v1) {return cons(v1);}
inline LObject list(const LObject& v1, const LObject& v2) {return cons(v1, cons(v2));}
inline LObject list(const LObject& v1, const LObject& v2, const LObject& v3) {
    return cons(v1, cons(v2, cons(v3)));
}
inline LObject list(const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4) {return cons(v1, cons(v2, cons(v3, cons(v4))));}
inline LObject list(const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5) {return cons(v1, cons(v2, cons(v3, cons(v4, cons(v5)))));}
inline LObject list(const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5, const LObject& v6) {return cons(v1, cons(v2, cons(v3, cons(v4, cons(v5, cons(v6))))));}
inline LObject list(const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5, const LObject& v6, const LObject& v7, const LObject& v8) {return cons(v1, cons(v2, cons(v3, cons(v4, cons(v5, cons(v6, cons(v7, cons(v8))))))));}

inline LObject seq(int symbol) { CUtiIDString str; str.setID(symbol); return  cons(str);}
inline LObject seq(int symbol, const LObject& v1) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1));}
inline LObject seq(int symbol, const LObject& v1, const LObject& v2) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1, cons(v2)));}
inline LObject seq(int symbol, const LObject& v1, const LObject& v2, const LObject& v3) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1, cons(v2, cons(v3))));}
inline LObject seq(int symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4)))));}
inline LObject seq(int symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4, cons(v5))))));}
inline LObject seq(int symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5, const LObject& v6) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4, cons(v5, cons(v6)))))));}
inline LObject seq(int symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5, const LObject& v6, const LObject& v7, const LObject& v8) {CUtiIDString str; str.setID(symbol); return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4, cons(v5, cons(v6, cons(v7, cons(v8)))))))));}

inline LObject seq(const CUtiIDString& symbol) { CUtiIDString str; str=symbol; return  cons(str);}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1) {CUtiIDString str; str=symbol; return  cons(str, cons(v1));}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1, const LObject& v2) {CUtiIDString str; str=symbol; return  cons(str, cons(v1, cons(v2)));}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1, const LObject& v2, const LObject& v3) {CUtiIDString str; str=symbol; return  cons(str, cons(v1, cons(v2, cons(v3))));}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4) {CUtiIDString str; str=symbol; return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4)))));}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5) {CUtiIDString str; str=symbol; return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4, cons(v5))))));}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5, const LObject& v6) {CUtiIDString str; str=symbol; return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4, cons(v5, cons(v6)))))));}
inline LObject seq(const CUtiIDString& symbol, const LObject& v1, const LObject& v2, const LObject& v3, const LObject& v4, const LObject& v5, const LObject& v6, const LObject& v7, const LObject& v8) {CUtiIDString str; str=symbol; return  cons(str, cons(v1, cons(v2, cons(v3, cons(v4, cons(v5, cons(v6, cons(v7, cons(v8)))))))));}



LObject append(const LObject & v2);
inline LObject eval(const LObject& obj) {
   return obj.m_data->l_eval();
}

LObject adjoin(const LObject& item, const LObject& list);
LObject adjoin(const LObject& item, const LObject& list, const LObject& keys);

bool member(const LObject& item, const LObject& list);
bool member(const LObject& item, const LObject& list, const LObject& keys);
LObject sort(const LObject& l1, const LObject& predicate);
LObject sort(const LObject& l1, const LObject& predicate, const LObject& keys);
LObject substitute(const LObject& newitem, const LObject& olditem, const LObject& sequence, const LObject& keys=LObject());
LObject remove(const LObject& olditem, const LObject& sequence, const LObject& keys= LObject());
LObject l_delete(const LObject& item,  const LObject& sequence, const LObject& keys= LObject());
int count(const LObject& olditem, const LObject& sequence, const LObject& keys);
LObject find(const LObject& olditem, const LObject& sequence, const LObject& keys= LObject());
LObject position(const LObject& olditem, const LObject& sequence, const LObject& keys= LObject());
int positioncpp(const LObject& olditem, const LObject& sequence, const LObject& keys= LObject());
LObject map(const LObject& resulttype, const LObject& function, const LObject& sequences);


inline LObject funcall(const LObject& data, const LObject& parameters){return cons(data, parameters ).eval();}
inline LObject funccall(const LObject& data, const LObject& p1){return list(data, p1).eval();}
inline LObject funccall(const LObject& data, const LObject& p1, const LObject& p2){return list(data, p1, p2).eval();}
inline bool evenp(const LObject& data) {return (data.toInt() % 2) == 0; }
inline bool oddp(const LObject& data) {return (data.toInt() % 2) == 1; }
inline double sin(const LObject& data) {return ::sin(data.toFloat());}
inline double cos(const LObject& data) {return ::cos(data.toFloat());}
inline double tan(const LObject& data) {return ::tan(data.toFloat());}
inline double asin(const LObject& data) {return ::asin(data.toFloat());}
inline double acos(const LObject& data) {return ::acos(data.toFloat());}
inline double atan(const LObject& data) {return ::atan(data.toFloat());}
inline double sinh(const LObject& data) {return ::sinh(data.toFloat());}
inline double cosh(const LObject& data) {return ::cosh(data.toFloat());}
inline double tanh(const LObject& data) {return ::tanh(data.toFloat());}
inline double asinh(const LObject& data) {return ::asinh(data.toFloat());}
inline double acosh(const LObject& data) {return ::acosh(data.toFloat());}
inline double atanh(const LObject& data) {return ::atanh(data.toFloat());}
inline LObject defvar(const LObject& varname, const LObject& objectinit) { LGlobalEnvironment::s_globalenvironment->createVariable(varname.getAtom(), objectinit); return varname;}
inline LObject l_if(const LObject& cond, const LObject& o1, const LObject &o2) {
    bool t = cond.eval().toBool();
  //  LGlobalEnvironment::s_globalenvironment->push();
    if (t) {
        return o1.eval();
    } else {
        return o2.eval();
    }
   // LGlobalEnvironment::s_globalenvironment->pop();
}

inline LObject when(const LObject& cond, const LObject& o1)
{
    bool t = cond.eval().toBool();
  //  LGlobalEnvironment::s_globalenvironment->push();
    if (t) {
        return o1.eval();
    };
    return LObject();
}
inline bool null(const LObject& obj) {return obj.isNil(); }
inline bool atom(const LObject& a) {return a.isAtom();}
inline bool listp(const LObject& a) {return a.isConList();}
bool l_and(const LObject& a1, const LObject& l);
bool l_or(const LObject& a1, const LObject& l);

inline bool l_not(const LObject& a1) {
    return !a1.eval().toBool();
}

LObject l_case(const LObject& cond);
LObject cond(const LObject& cond);
inline bool equal(const LObject& c1, const LObject& c2) {return c1 == c2;};
LObject defun(const LObject& name, const LObject& parameter, const LObject& body);
//also for loop
LObject loop(const LObject& sexpressions);
LObject l_do(const LObject& evalop, const LObject& conditions, const LObject& sexpressions);
LObject dotimes(const LObject& amount, const LObject& sexpressions);
LObject dolist(const LObject& amount, const LObject& sexpressions);
LObject lambda(const LObject&parameters, const LObject& body);
LObject let(const LObject& variables, const LObject& sexpressions);
LObject let_star(const LObject& variables, const LObject& sexpressions);
LObject block(const LObject& name, const LObject& sexpressions);
LObject l_return(const LObject& valuelist);
LObject l_return_from(const LObject& name, const LObject& valuelist);
LObject plus(const LObject& f, const LObject& rest);
LObject minus(const LObject& f, const LObject& rest);
LObject mul(const LObject& f, const LObject& rest);
LObject div(const LObject& f, const LObject& rest);
LObject mod(const LObject& f, const LObject& rest);

LObject max(const LObject& f, const LObject& rest);
LObject min(const LObject& f, const LObject& rest);
LObject abs(const LObject& f);
LObject round(const LObject& f);
LObject trunc(const LObject& f);
LObject random(const LObject& f);
LObject setf(const LObject& varname, const LObject& exp);
LObject executequote(const LObject& exp);
LObject quote(const LObject& exp);
bool equal_numeric(const LObject& f1,const LObject& f2);
bool less(const LObject& f1,const LObject& f2);
bool greater(const LObject& f1,const LObject& f2);
bool eless(const LObject& f1,const LObject& f2);
bool egreater(const LObject& f1,const LObject& f2);
LObject assoc(const LObject& f1,const LObject& f2);
LObject single_plus(const LObject& f1, const LObject& f2);
LObject single_minus(const LObject& f1, const LObject& f2);
LObject single_mul(const LObject& f1, const LObject& f2);
LObject single_div(const LObject& f1, const LObject& f2);
LObject single_mod(const LObject& f1, const LObject& f2);

inline LObject inc_one(const LObject& data) {return single_plus(data, LObject(1)); }
inline LObject dec_one(const LObject& data) {return single_plus(data, LObject(-1));  }
inline bool eql(const LObject& d1, const LObject& d2) {return d1.m_data.get() == d2.m_data.get();}
inline LObject incf(const LObject& data, const LObject& d1) {return single_plus(data, d1);}
inline LObject decf(const LObject& data, const LObject& d2) {return single_minus(data, d2); }
inline LObject type_of(const LObject& obj) {return obj.m_data->l_typeof();}

inline LObject typep(const LObject& obj, const LObject& typeatom) {return obj.m_data->l_typep(typeatom.getAtom().getID());}

class LispErrorContainer : public ReferenceUtiOB
{
public:
    std::map<LCell*, CIVector2> m_element_to_line;
    CUtiString m_filename;
    CUtiString createError(LCell* cell, const CUtiString& name);

    bool hasErrorInfo(LCell* cell);
    std::map<LCell*, CIVector2>::iterator getErrorRecursive(LCell* cell);

    LObject take(const LObject& obj, const CIVector2 &pos);
    void replacedBy(const LObject& old, const LObject &newobj);
};



void parseStringAll(const CUtiString& data, LispErrorContainer* errorcontainer, std::vector<LObject>& resobj);
LObject parseString(const CUtiString& data, LispErrorContainer* errorcontainer=NULL);

#define LISPSTR(s) Lisp::parseString(UTISTR(s))

class LArray : public LBaseArray
{
public:
    Array< uti_refptr<LCell> > m_value;
    LArray(int size) {m_type = tctArray;m_value.setSize(size);}
    virtual int l_typeof() const {return LISP_vector;}
    virtual bool l_typep(int type) const {return type == LISP_vector;}

};

class LMultiDimensionArray : public LBaseArray
{
public:
    Array<int> m_dimensions;
    Array< uti_refptr<LCell> > m_value;
    LMultiDimensionArray(int size) {m_type =tctMultiArray; m_value.setSize(size);}

};


class LDynamicArray : public LBaseArray
{
public:
    DynArray1D< uti_refptr<LCell> > m_value;
    LDynamicArray(){m_type= tctDynamicArray;}

};




LObject::LObject(){ m_data = new LNil();updateDebug();}
LObject::LObject(int value) {m_data = new LInteger(value);updateDebug();}
LObject::LObject(double value) {m_data = new LFloat(value);updateDebug();}
LObject::LObject(float value) {m_data = new LFloat(value);updateDebug();}
LObject::LObject(bool b) { if (b) { m_data = new LTrue();} else { m_data = new LNil();};updateDebug();}
LObject::LObject(const CUtiString& str) {m_data = new LString(str);updateDebug();}
LObject::LObject(const CUtiIDString& str) {m_data = new LAtom(str);updateDebug();}
LObject::LObject(const LObject& str) {m_data = str.m_data;
                                  #ifdef DEBUGLISP
                                      updateDebug();//m_content = str.m_content;
                                  #endif

                                     }



LObject::LObject(LString* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LString* cell) {m_data = const_cast<LString*>(cell);updateDebug();}
LObject::LObject(LInteger* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LInteger* cell) {m_data = const_cast<LInteger*>(cell);updateDebug();}
LObject::LObject(LFloat* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LFloat* cell) {m_data = const_cast<LFloat*>(cell);updateDebug();}
LObject::LObject(LList* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LList* cell) {m_data = const_cast<LList*>(cell);updateDebug();}
LObject::LObject(LCons* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LCons* cell) {m_data = const_cast<LCons*>(cell);updateDebug();}
LObject::LObject(LBaseArray* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LBaseArray* cell) {m_data = const_cast<LBaseArray*>(cell);updateDebug();}


LObject::LObject(LAtom* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LAtom* cell) {m_data = const_cast<LAtom*>(cell);updateDebug();}
LObject::LObject(LTrue* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LTrue* cell) {m_data = const_cast<LTrue*>(cell);updateDebug();}
LObject::LObject(LNil* cell) {m_data = cell;updateDebug();}
LObject::LObject(const LNil* cell) {m_data = const_cast<LNil*>(cell);updateDebug();}

LObject& LObject::operator=(int value) {m_data = new LInteger(value);updateDebug(); return *this;}
LObject& LObject::operator=(float value) {m_data = new LFloat(value);updateDebug();return *this;}
LObject& LObject::operator=(double value) {m_data = new LFloat(value);updateDebug();return *this;}
LObject& LObject::operator=(bool b) {if (b) { m_data = new LTrue();} else { m_data = new LNil();};updateDebug();return *this;}
LObject& LObject::operator=(const CUtiString& str) {m_data = new LString(str);updateDebug();return *this;}
LObject& LObject::operator=(const CUtiIDString& str) {m_data = new LAtom(str);updateDebug();return *this;}
LObject& LObject::operator=(const LObject& str) {m_data = str.m_data;
                                             #ifdef DEBUGLISP
                                              updateDebug();  // m_content = str.m_content;
                                             #endif
                                                 return *this;}

bool LObject::isConList() const
{
    return m_data->m_type == tctCons;
}
bool LObject::isNil() const
{
    return m_data->m_type == tctNil;
}
void LObject::setNil()
{
    m_data = new LNil();
}
bool LObject::isAtom() const
{
    return m_data->m_type == tctAtom;
}


bool LObject::isString() const
{
    return m_data->m_type == tctString;

}

bool LObject::isThisAtom(const CUtiIDString& str) const
{
    if (!isAtom()) return false;

    return getAtom() == str;

}
CUtiIDString LObject::getAtom() const
{
    if (!isAtom()) { throw LispException(*this, UTISTR("Atom is here expected"));}
    return check_cast<LAtom*>(m_data.get())->m_value;
}

inline bool LObject::isBool() const
{
    switch (m_data->m_type) {
    case tctTrue: return true;
    case tctNil: return true;
    default: {

        return false;
    }; break;

    };
}


inline bool LObject::toBool() const
{
    switch (m_data->m_type) {
    case tctTrue: return true;
    case tctNil: return false;
    default: {
        throw LispException(*this, UTISTR("Cannot convert to bool"));
        return false;
    }; break;

    };
}


inline CUtiString LObject::toString() const
{
    if (m_data->m_type == tctString) {

        return check_cast<LString*>(m_data.get())->m_value;
    }  else {
        throw LispException(*this, UTISTR("Cannot convert to string"));
        return EMPTYSTRING;

    }
}

inline double LObject::toFloat() const
{
    if (m_data->m_type == tctInteger) {

        return check_cast<LInteger*>(m_data.get())->m_value;
    } else if (m_data->m_type == tctFloat) {

        return check_cast<LFloat*>(m_data.get())->m_value;
    } else {
        throw LispException(*this, UTISTR("Cannot convert to float"));
        return 0;

    }

}  /*
*/
inline int LObject::toInt() const
{
    if (m_data->m_type == tctInteger) {

        return check_cast<LInteger*>(m_data.get())->m_value;
    } else {
        throw LispException(*this, UTISTR("Cannot convert to int"));
        return false;

    }
}

}

#endif // UTILISP_H
